import os

from numpy.testing import assert_array_equal

from glue.core.coordinates import WCSCoordinates
from glue.core import data_factories as df
from glue.tests.helpers import requires_pyavm, requires_pil_or_skimage, make_file

DATA = os.path.join(os.path.dirname(__file__), 'data')


@requires_pil_or_skimage
def test_grey_png_loader():
    # Greyscale PNG
    data = b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x02\x00\x00\x00\x02\x08\x00\x00\x00\x00W\xddR\xf8\x00\x00\x00\x0eIDATx\x9ccdddab\x04\x00\x00&\x00\x0b\x8e`\xe7A\x00\x00\x00\x00IEND\xaeB`\x82'
    with make_file(data, '.png') as fname:
        d = df.load_data(fname)
        assert df.find_factory(fname) is df.img_data
    assert_array_equal(d['PRIMARY'], [[3, 4], [1, 2]])


@requires_pil_or_skimage
def test_color_png_loader():
    # Colorscale PNG
    data = b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x02\x00\x00\x00\x02\x08\x02\x00\x00\x00\xfd\xd4\x9as\x00\x00\x00\x15IDAT\x08\xd7\x05\xc1\x01\x01\x00\x00\x00\x80\x10\xffO\x17B\x14\x1a!\xec\x04\xfc\xf2!Q\\\x00\x00\x00\x00IEND\xaeB`\x82'
    with make_file(data, '.png') as fname:
        d = df.load_data(fname)
        assert df.find_factory(fname) is df.img_data
    assert_array_equal(d['red'], [[255, 0], [255, 0]])
    assert_array_equal(d['green'], [[255, 0], [0, 255]])
    assert_array_equal(d['blue'], [[0, 255], [0, 0]])


@requires_pyavm
@requires_pil_or_skimage
def test_avm():
    data = df.load_data(os.path.join(DATA, 'ssc2006-16a1_Ti.jpg'))
    assert isinstance(data.coords, WCSCoordinates)



import pytest

from glue.core.state_objects import State
from ..elements import (FormElement, NumberElement, ChoiceElement,
                        FloatElement, TextBoxElement)


def get_value(element):

    prefix, widget_cls, property = element.ui_and_state()

    class TemporaryState(State):
        a = property

    temp = TemporaryState()
    return temp.a


class TestFormElements(object):

    def test_number_default_value(self):
        e = FormElement.auto((0, 100, 30))
        assert get_value(e) == 30

    def test_number_float(self):
        e = FormElement.auto((0.0, 1.0, 0.3))
        assert get_value(e) == 0.3

    def test_number_list(self):
        e = FormElement.auto([0, 10])
        assert isinstance(e, NumberElement)

    def test_choice_list(self):
        e = FormElement.auto(['a', 'b'])
        assert isinstance(e, ChoiceElement)

    def test_choice_tuple(self):
        e = FormElement.auto(('a', 'b'))
        assert isinstance(e, ChoiceElement)

    def test_float(self):
        e = FormElement.auto(1.2)
        assert isinstance(e, FloatElement)

        e = FormElement.auto(2)
        assert isinstance(e, FloatElement)
        assert get_value(e) == 2

    def test_textbox(self):
        e = FormElement.auto('_str')
        assert isinstance(e, TextBoxElement)
        assert get_value(e) == 'str'

    def test_recognizes_subsubclasses(self):

        class SubClassFormElement(TextBoxElement):
            @classmethod
            def recognizes(cls, params):
                return params == 'specific_class'

        e = FormElement.auto('specific_class')
        assert isinstance(e, SubClassFormElement)

    def test_unrecognized(self):
        with pytest.raises(ValueError):
            FormElement.auto(None)



import pytest

from numpy.testing import assert_equal
from pandas import DataFrame

from glue.core import Data, DataCollection


def test_translator_data_roundtrip():

    # Case where we start from a Dataframe, convert to a glue data object,
    # and convert back to a DataFrame.

    df = DataFrame()
    df['a'] = [3, 5, 6, 7]
    df['b'] = [1.5, 2.2, 1.3, 3.3]
    df['c'] = ['r', 'd', 'w', 'q']

    dc = DataCollection()
    dc['dataframe'] = df

    data = dc['dataframe']
    assert isinstance(data, Data)
    assert data.main_components == ['a', 'b', 'c']
    assert_equal(data['a'], [3, 5, 6, 7])
    assert_equal(data['b'], [1.5, 2.2, 1.3, 3.3])
    assert_equal(data['c'], ['r', 'd', 'w', 'q'])

    df2 = data.get_object()
    assert_equal(list(df2.columns), ['a', 'b', 'c'])
    assert_equal(df2['a'].values, [3, 5, 6, 7])
    assert_equal(df2['b'].values, [1.5, 2.2, 1.3, 3.3])
    assert_equal(df2['c'].values, ['r', 'd', 'w', 'q'])


def test_translator_from_data():

    # Case where the initial data object wasn't originally created from a
    # DataFrame.

    data = Data()
    data['a'] = [3, 5, 6, 7]
    data['b'] = [1.5, 2.2, 1.3, 3.3]
    data['c'] = ['r', 'd', 'w', 'q']

    with pytest.raises(ValueError) as exc:
        df = data.get_object()
    assert exc.value.args[0] == ('Specify the object class to use with cls= - supported '
                                 'classes are:\n\n* pandas.core.frame.DataFrame')

    df = data.get_object(cls=DataFrame)
    assert_equal(list(df.columns), ['a', 'b', 'c'])
    assert_equal(df['a'].values, [3, 5, 6, 7])
    assert_equal(df['b'].values, [1.5, 2.2, 1.3, 3.3])
    assert_equal(df['c'].values, ['r', 'd', 'w', 'q'])


def test_translator_from_subset():

    # Case where we convert a subset to a DataFrame

    data = Data()
    data['a'] = [3, 5, 6, 7]
    data['b'] = [1.5, 2.2, 1.3, 3.3]
    data['c'] = ['r', 'd', 'w', 'q']

    dc = DataCollection([data])
    dc.new_subset_group(label='test subset', subset_state=data.id['b'] > 2)

    df = data.get_subset_object(cls=DataFrame)
    assert_equal(list(df.columns), ['a', 'b', 'c'])
    assert_equal(df['a'].values, [5, 7])
    assert_equal(df['b'].values, [2.2, 3.3])
    assert_equal(df['c'].values, ['d', 'q'])


def test_translator_from_data_with_derived():

    # Case where we convert a subset to a DataFrame

    data = Data()
    data['a'] = [3, 5, 6, 7]
    data['b'] = data.id['a'] + 1

    dc = DataCollection([data])
    dc.new_subset_group(label='test subset', subset_state=data.id['b'] > 2)

    df = data.get_object(cls=DataFrame)
    assert_equal(list(df.columns), ['a', 'b'])
    assert_equal(df['a'].values, [3, 5, 6, 7])
    assert_equal(df['b'].values, [4, 6, 7, 8])


# TODO: Case where we convert from a BaseData object



from echo import keep_in_sync, CallbackProperty
from glue.core.layer_artist import LayerArtistBase
from glue.viewers.common.state import LayerState
from glue.core.message import LayerArtistVisibilityMessage

__all__ = ['LayerArtist']


class LayerArtist(LayerArtistBase):

    zorder = CallbackProperty()
    visible = CallbackProperty()

    _layer_state_cls = LayerState

    _python_exporter = None

    def __init__(self, viewer_state, layer_state=None, layer=None):

        super(LayerArtist, self).__init__(layer)

        self._viewer_state = viewer_state

        self.layer = layer or layer_state.layer
        self.state = layer_state or self._layer_state_cls(viewer_state=viewer_state,
                                                          layer=self.layer)

        if self.state not in self._viewer_state.layers:
            self._viewer_state.layers.append(self.state)

        self.zorder = self.state.zorder
        self.visible = self.state.visible

        self._sync_zorder = keep_in_sync(self, 'zorder', self.state, 'zorder')
        self._sync_visible = keep_in_sync(self, 'visible', self.state, 'visible')

        self.state.add_callback('visible', self._on_visibility_change)

        self._reset_cache()

    def _reset_cache(self):
        self._last_viewer_state = {}
        self._last_layer_state = {}

    def _on_visibility_change(self, *args):
        if self.state.layer is not None and self.state.layer.hub is not None:
            self.state.layer.hub.broadcast(LayerArtistVisibilityMessage(self))

    def __gluestate__(self, context):
        return dict(state=context.id(self.state))

    def pop_changed_properties(self):
        """
        Return the names of properties on the viewer and layer state classes
        that have changed since the last call.

        Note that calling this method updates the underlying cache, so if it is
        called immediately after being called a first time, it will return an
        empty set the second time.
        """

        # Figure out which attributes are different from before. Ideally we
        # shouldn't need this but some methods in layer artists are called
        # multiple times if an attribute is changed due to x_att changing then
        # hist_x_min, hist_x_max, etc. This method is called pop because it
        # returns the list of changed properties and resets the cache.

        changed = set()

        for key, value in self._viewer_state.as_dict().items():
            if value != self._last_viewer_state.get(key, None):
                changed.add(key)

        for key, value in self.state.as_dict().items():
            if value != self._last_layer_state.get(key, None):
                changed.add(key)

        self._last_viewer_state.update(self._viewer_state.as_dict())
        self._last_layer_state.update(self.state.as_dict())

        return changed



from glue.viewers.common.python_export import code, serialize_options


def python_export_image_layer(layer, *args):

    if not layer.enabled or not layer.visible:
        return [], None

    script = ""
    imports = ["from glue.viewers.image.state import get_sliced_data_maker"]
    imports += ["import matplotlib.patches as mpatches"]

    slices, agg_func, transpose = layer._viewer_state.numpy_slice_aggregation_transpose

    # TODO: implement aggregation, ignore for now

    script += "# Define a function that will get a fixed resolution buffer\n"

    options = {'data': code('layer_data'),
               'x_axis': layer._viewer_state.x_att.axis,
               'y_axis': layer._viewer_state.y_att.axis,
               'slices': slices,
               'target_cid': code("layer_data.id['{0}']".format(layer.state.attribute))}

    if transpose:
        options['transpose'] = True

    script += "array_maker = get_sliced_data_maker({0})\n\n".format(serialize_options(options))

    script += "composite.allocate('{0}')\n".format(layer.uuid)

    if layer._viewer_state.color_mode == 'Colormaps':
        color = code('plt.cm.' + layer.state.cmap.name)
    else:
        color = layer.state.color

    options = dict(array=code('array_maker'),
                   clim=(layer.state.v_min, layer.state.v_max),
                   visible=layer.state.visible,
                   zorder=layer.state.zorder,
                   color=color,
                   contrast=layer.state.contrast,
                   bias=layer.state.bias,
                   alpha=layer.state.alpha,
                   stretch=layer.state.stretch)

    script += "composite.set('{0}', {1})\n".format(layer.uuid, serialize_options(options))

    if layer._viewer_state.color_mode == 'Colormaps':
        imports += ["from glue.utils.matplotlib import ColormapPatchHandler"]
        script += "handle = mpatches.Patch(color='{0}')\n".format(layer.state.color)
        script += "handler = ColormapPatchHandler(" + code('plt.cm.' + layer.state.cmap.name) + ")\n"

        script += "legend_handles.append(handle)\n"
        script += "legend_handler_dict[handle] = handler\n"
        script += "legend_labels.append(layer_data.label)\n"
    else:
        options = dict(color=layer.state.color,
                       alpha=layer.state.alpha)
        script += "handle = mpatches.Patch({0})\n".format(serialize_options(options))
        script += "legend_handles.append(handle)\n"
        script += "legend_labels.append(layer_data.label)\n"

    script += "\n"

    return imports, script.strip()


def python_export_image_subset_layer(layer, *args):

    if not layer.enabled or not layer.visible:
        return [], None

    script = ""
    imports = ["from glue.viewers.image.state import get_sliced_data_maker"]

    slices, agg_func, transpose = layer._viewer_state.numpy_slice_aggregation_transpose

    # TODO: implement aggregation, ignore for now

    script += "# Define a function that will get a fixed resolution buffer of the mask\n"

    options = {'data': code('layer_data'),
               'x_axis': layer._viewer_state.x_att.axis,
               'y_axis': layer._viewer_state.y_att.axis,
               'slices': slices}

    if transpose:
        options['transpose'] = True

    script += "array_maker = get_sliced_data_maker({0})\n\n".format(serialize_options(options))

    script += "\n\n"

    options = dict(origin='lower', interpolation='nearest', color=layer.state.color,
                   vmin=0, vmax=1, aspect=layer._viewer_state.aspect,
                   zorder=layer.state.zorder, alpha=layer.state.alpha)

    script += "imshow(ax, {0}, {1})\n".format(code('array_maker'), serialize_options(options))

    # legend
    options = dict(color=layer.state.color,
                   alpha=layer.state.alpha)
    script += "handle = mpatches.Patch({0})\n".format(serialize_options(options))
    script += "legend_handles.append(handle)\n"
    script += "legend_labels.append(layer_data.label)\n"
    script += "\n"

    return imports, script.strip()



from qtpy.QtCore import Qt
from qtpy import QtWidgets
from glue.core.qt.layer_artist_model import QtLayerArtistContainer, LayerArtistWidget
from glue.utils.qt import set_cursor, messagebox_on_error
from glue.core.qt.dialogs import warn
from glue.utils.noconflict import classmaker
from glue.config import viewer_tool
from glue.viewers.common.qt.base_widget import BaseQtViewerWidget
from glue.viewers.common.tool import SimpleToolMenu
from glue.viewers.common.qt.toolbar import BasicToolbar
from glue.viewers.common.viewer import Viewer
from glue.viewers.common.utils import get_viewer_tools

__all__ = ['DataViewer', 'get_viewer_tools']


class ToolbarInitializer(object):
    """
    This is a meta-class which ensures that initialize_toolbar is always called
    on DataViewer instances and sub-class instances after all the __init__ code
    has been executed. We need to do this, because often the toolbar can only
    be initialized after everything else (e.g. canvas, etc.) has been set up,
    so we can't do it in DataViewer.__init__.
    """

    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj.initialize_toolbar()
        return obj


@viewer_tool
class SaveTool(SimpleToolMenu):
    """
    A generic 'save/export' tool that plugins can register new save/export tools
    with.

    To register a new save/export option, add an entry to the viewer
    ``subtools['save']`` list.
    """
    tool_id = 'save'
    icon = 'glue_filesave'
    tool_tip = 'Save/export the plot'


# Note: we need to use classmaker here because otherwise we run into issues when
# trying to use the meta-class with the Qt class.
class DataViewer(Viewer, BaseQtViewerWidget,
                 metaclass=classmaker(left_metas=(ToolbarInitializer,))):
    """
    Base class for all Qt DataViewer widgets.

    This defines a minimal interface, and implements the following::

       * An automatic call to unregister on window close
       * Drag and drop support for adding data
    """

    _layer_artist_container_cls = QtLayerArtistContainer
    _layer_style_widget_cls = None

    _toolbar_cls = BasicToolbar

    # This defines the mouse mode to be used when no toolbar modes are active
    _default_mouse_mode_cls = None

    inherit_tools = True
    tools = ['save']
    subtools = {'save': []}

    _close_on_last_layer_removed = True

    _options_cls = None

    large_data_size = None

    def __init__(self, session, state=None, parent=None):
        """
        :type session: :class:`~glue.core.session.Session`
        """

        BaseQtViewerWidget.__init__(self, parent)
        Viewer.__init__(self, session, state=state)

        self._view = LayerArtistWidget(layer_style_widget_cls=self._layer_style_widget_cls,
                                       hub=session.hub)
        self._view.layer_list.setModel(self._layer_artist_container.model)

        # Set up the options widget, which will include options that control the
        # viewer state
        if self._options_cls is None:
            self.options = QtWidgets.QWidget()
        else:
            self.options = self._options_cls(viewer_state=self.state,
                                             session=session)

        self._tb_vis = {}  # store whether toolbars are enabled
        self.toolbar = None
        self._toolbars = []
        self._warn_close = True

        # close window when last plot layer deleted
        if self._close_on_last_layer_removed:
            self._layer_artist_container.on_empty(self._close_nowarn)
        self._layer_artist_container.on_changed(self.update_window_title)

        self.update_window_title()

    @property
    def selected_layer(self):
        return self._view.layer_list.current_artist()

    @set_cursor(Qt.WaitCursor)
    def apply_roi(self, roi):
        pass

    def warn(self, message, *args, **kwargs):
        return warn(message, *args, **kwargs)

    def _close_nowarn(self):
        return self.close(warn=False)

    def closeEvent(self, event):
        super(DataViewer, self).closeEvent(event)
        Viewer.cleanup(self)
        # We tell the toolbar to do cleanup to make sure we get rid of any
        # circular references
        if self.toolbar:
            self.toolbar.cleanup()

    def layer_view(self):
        return self._view

    def addToolBar(self, tb):
        super(DataViewer, self).addToolBar(tb)
        self._toolbars.append(tb)
        self._tb_vis[tb] = True

    def initialize_toolbar(self):

        from glue.config import viewer_tool

        self.toolbar = self._toolbar_cls(self, default_mouse_mode_cls=self._default_mouse_mode_cls)

        # Need to include tools and subtools declared by parent classes unless
        # specified otherwise
        tool_ids, subtool_ids = get_viewer_tools(self.__class__)

        for tool_id in tool_ids:
            mode_cls = viewer_tool.members[tool_id]
            if tool_id in subtool_ids:
                subtools = []
                for subtool_id in subtool_ids[tool_id]:
                    subtools.append(viewer_tool.members[subtool_id](self))
                mode = mode_cls(self, subtools=subtools)
            else:
                mode = mode_cls(self)
            self.toolbar.add_tool(mode)

        self.addToolBar(self.toolbar)

        self.toolbar_added.emit()

    def show_toolbars(self):
        """
        Re-enable any toolbars that were hidden with `hide_toolbars()`

        Does not re-enable toolbars that were hidden by other means
        """
        for tb in self._toolbars:
            if self._tb_vis.get(tb, False):
                tb.setEnabled(True)

    def hide_toolbars(self):
        """
        Disable all the toolbars in the viewer.

        This action can be reversed by calling `show_toolbars()`
        """
        for tb in self._toolbars:
            self._tb_vis[tb] = self._tb_vis.get(tb, False) or tb.isVisible()
            tb.setEnabled(False)

    def set_focus(self, state):
        super(DataViewer, self).set_focus(state)
        if state:
            self.show_toolbars()
        else:
            self.hide_toolbars()

    def __gluestate__(self, context):
        state = Viewer.__gluestate__(self, context)
        state['size'] = self.viewer_size
        state['pos'] = self.position
        state['_protocol'] = 1
        return state

    def update_viewer_state(rec, context):
        pass

    @classmethod
    def __setgluestate__(cls, rec, context):

        if rec.get('_protocol', 0) < 1:
            cls.update_viewer_state(rec, context)

        viewer = super(DataViewer, cls).__setgluestate__(rec, context)

        viewer.viewer_size = rec['size']
        x, y = rec['pos']
        viewer.move(x=x, y=y)

        return viewer

    @messagebox_on_error("Failed to add data")
    def add_data(self, data):
        return super(DataViewer, self).add_data(data)

    @messagebox_on_error("Failed to add subset")
    def add_subset(self, subset):
        return super(DataViewer, self).add_subset(subset)



from matplotlib.backends.backend_qt5 import NavigationToolbar2QT

from glue.config import viewer_tool
from glue.viewers.common.tool import CheckableTool, Tool


__all__ = ['MatplotlibTool', 'MatplotlibCheckableTool', 'HomeTool', 'SaveTool',
           'PanTool', 'ZoomTool']


def _ensure_mpl_nav(viewer):
    # Set up virtual Matplotlib navigation toolbar (don't show it)
    if not hasattr(viewer, '_mpl_nav'):
        viewer._mpl_nav = NavigationToolbar2QT(viewer.central_widget.canvas, viewer)
        viewer._mpl_nav.hide()


def _cleanup_mpl_nav(viewer):
    if getattr(viewer, '_mpl_nav', None) is not None:
        viewer._mpl_nav.setParent(None)
        try:
            viewer._mpl_nav.parent = None
        except AttributeError:
            pass
        viewer._mpl_nav = None


class MatplotlibTool(Tool):

    def __init__(self, viewer=None):
        super(MatplotlibTool, self).__init__(viewer=viewer)
        _ensure_mpl_nav(viewer)

    def close(self):
        _cleanup_mpl_nav(self.viewer)
        super(MatplotlibTool, self).close()


class MatplotlibCheckableTool(CheckableTool):

    def __init__(self, viewer=None):
        super(MatplotlibCheckableTool, self).__init__(viewer=viewer)
        _ensure_mpl_nav(viewer)

    def close(self):
        _cleanup_mpl_nav(self.viewer)
        super(MatplotlibCheckableTool, self).close()


@viewer_tool
class HomeTool(MatplotlibTool):

    tool_id = 'mpl:home'
    icon = 'glue_home'
    action_text = 'Home'
    tool_tip = 'Reset original zoom'
    shortcut = 'H'

    def activate(self):
        if hasattr(self.viewer, 'state') and hasattr(self.viewer.state, 'reset_limits'):
            self.viewer.state.reset_limits()
        else:
            self.viewer._mpl_nav.home()


@viewer_tool
class SaveTool(MatplotlibTool):

    tool_id = 'mpl:save'
    icon = 'glue_filesave'
    action_text = 'Save plot to file'
    tool_tip = 'Save the figure'

    def activate(self):
        self.viewer._mpl_nav.save_figure()


@viewer_tool
class PanTool(MatplotlibCheckableTool):

    tool_id = 'mpl:pan'
    icon = 'glue_move'
    action_text = 'Pan'
    tool_tip = 'Pan axes with left mouse, zoom with right'
    shortcut = 'M'

    def activate(self):
        self.viewer._mpl_nav.pan()

    def deactivate(self):
        if hasattr(self.viewer, '_mpl_nav'):
            self.viewer._mpl_nav.pan()


@viewer_tool
class ZoomTool(MatplotlibCheckableTool):

    tool_id = 'mpl:zoom'
    icon = 'glue_zoom_to_rect'
    action_text = 'Zoom'
    tool_tip = 'Zoom to rectangle'
    shortcut = 'Z'

    def activate(self):
        self.viewer._mpl_nav.zoom()

    def deactivate(self):
        if hasattr(self.viewer, '_mpl_nav'):
            self.viewer._mpl_nav.zoom()



# pylint: disable=I0011,W0613,W0201,W0212,E1101,E1103

import operator

import pytest
import numpy as np
from unittest.mock import MagicMock

from glue import core
from glue.tests.helpers import requires_astropy

from ..coordinates import Coordinates
from ..component import (Component, DerivedComponent, CoordinateComponent,
                         CategoricalComponent)
from ..component_id import ComponentID
from ..data import Data
from ..parse import ParsedCommand, ParsedComponentLink
from ..data_collection import DataCollection

VIEWS = (np.s_[:], np.s_[1], np.s_[::-1], np.s_[0, :])


class TestComponent(object):

    def setup_method(self, method):
        self.data = MagicMock()
        self.data.shape = [1, 2]
        self.component = Component(self.data)

    def test_data(self):
        assert self.component.data is self.data

    def test_shape(self):
        assert self.component.shape is self.data.shape

    def test_ndim(self):
        assert self.component.ndim is len(self.data.shape)


class TestComponentID(object):

    def setup_method(self, method):
        self.cid = ComponentID('test')

    def test_label(self):
        assert self.cid.label == 'test'

    def test_str(self):
        """ str should return """
        str(self.cid)

    def test_repr(self):
        """ str should return """
        repr(self.cid)


class TestDerivedComponent(object):

    def setup_method(self, method):
        data = MagicMock()
        link = MagicMock()
        self.cid = DerivedComponent(data, link)
        self.link = link
        self.data = data

    def test_data(self):
        """ data property should wrap to links compute method """
        self.cid.data
        self.link.compute.assert_called_once_with(self.data)

    def test_link(self):
        assert self.cid.link == self.link


class TestCategoricalComponent(object):

    def setup_method(self, method):
        self.list_data = ['a', 'a', 'b', 'b']
        self.array_data = np.array(self.list_data)

    def test_autodetection(self):
        assert Component.autotyped(self.array_data).categorical
        assert Component.autotyped(self.list_data).categorical

        x = np.array([True, False, True, False])
        assert not Component.autotyped(x).categorical

        x = np.array([1, 2, 3, 4])
        assert not Component.autotyped(x).categorical

        x = np.array(['1', '2', '3', '4'])
        assert not Component.autotyped(x).categorical

        d = Data(x=['a', 'b', 'c'])
        assert d.get_component('x').categorical

    def test_basic_properties(self):
        data = ['a', 'b', 'c', 'b', 'b', 'c', 'a', 'c', 'd']
        cat_comp = CategoricalComponent(data)
        np.testing.assert_equal(cat_comp.data, data)
        np.testing.assert_equal(cat_comp.labels, data)
        np.testing.assert_equal(cat_comp.codes, [0, 1, 2, 1, 1, 2, 0, 2, 3])
        np.testing.assert_equal(cat_comp.categories, ['a', 'b', 'c', 'd'])

    def test_accepts_numpy(self):
        cat_comp = CategoricalComponent(self.array_data)
        assert cat_comp.data.shape == (4,)

    def test_accepts_pandas(self):
        # Regression test for a bug that caused read-only
        # pandas DataFrame columns to cause issues
        from pandas import DataFrame
        df = DataFrame()
        df['a'] = ['a', 'b', 'c']
        print(df['a'].fillna(''), type)
        CategoricalComponent(df['a'].fillna(''))

    def test_accepts_list(self):
        """Should accept a list and convert to numpy!"""
        cat_comp = CategoricalComponent(self.list_data)
        np.testing.assert_equal(cat_comp.data, self.array_data)

    def test_multi_nans(self):
        cat_comp = CategoricalComponent(['', '', 'a', 'b', 'c', 'zanthia'])
        np.testing.assert_equal(cat_comp.codes,
                                np.array([0, 0, 1, 2, 3, 4]))
        np.testing.assert_equal(cat_comp.categories,
                                np.asarray(['', 'a', 'b', 'c', 'zanthia']))

    def test_calculate_grouping(self):
        cat_comp = CategoricalComponent(self.array_data)
        np.testing.assert_equal(cat_comp.categories, np.asarray(['a', 'b']))
        np.testing.assert_equal(cat_comp.codes, np.array([0, 0, 1, 1]))
        assert cat_comp.codes.dtype == np.float

    def test_accepts_provided_grouping(self):
        ncategories = ['b', 'c']
        cat_data = list('aaabbbcccddd')
        cat_comp = CategoricalComponent(cat_data, categories=ncategories)

        assert cat_comp.categories == ncategories
        assert np.all(np.isnan(cat_comp.codes[:3]))
        assert np.all(cat_comp.codes[3:6] == 0)
        assert np.all(cat_comp.codes[6:9] == 1)
        assert np.all(np.isnan(cat_comp.codes[9:]))

    def test_uniform_jitter(self):
        cat_comp = CategoricalComponent(self.array_data)
        second_comp = CategoricalComponent(self.array_data)
        cat_comp.jitter(method='uniform')
        assert np.all(cat_comp.codes != second_comp.codes), "Didn't jitter data!"

    def test_unjitter_data(self):
        cat_comp = CategoricalComponent(self.array_data)
        second_comp = CategoricalComponent(self.array_data)

        cat_comp.jitter(method='uniform')
        delta = np.abs(cat_comp.codes - second_comp.codes).sum()
        assert delta > 0

        cat_comp.jitter(method=None)
        np.testing.assert_equal(cat_comp.codes,
                                second_comp.codes,
                                "Didn't un-jitter data!")

    def test_jitter_on_init(self):
        cat_comp = CategoricalComponent(self.array_data, jitter='uniform')
        second_comp = CategoricalComponent(self.array_data)
        assert np.all(cat_comp.codes != second_comp.codes)

    def test_object_dtype(self):
        d = np.array([1, 3, 3, 1, 'a', 'b', 'a'], dtype=object)
        c = CategoricalComponent(d)

        np.testing.assert_array_equal(c.categories,
                                      np.array([1, 3, 'a', 'b'], dtype=object))
        np.testing.assert_array_equal(c.codes, [0, 1, 1, 0, 2, 3, 2])

    def test_valueerror_on_bad_jitter(self):

        with pytest.raises(ValueError):
            cat_comp = CategoricalComponent(self.array_data)
            cat_comp.jitter(method='this will never be a jitter method')


class TestCoordinateComponent(object):

    def setup_method(self, method):

        class TestCoords(Coordinates):

            def __init__(self):
                super().__init__(pixel_n_dim=3, world_n_dim=3)

            def pixel_to_world_values(self, *args):
                return [a * (i + 1) for i, a in enumerate(args)]

            def world_to_pixel_values(self, *args):
                return [a / (i + 1) for i, a in enumerate(args)]

            @property
            def axis_correlation_matrix(self):
                return np.identity(3).astype(bool)

        data = core.Data()
        data.add_component(Component(np.zeros((3, 3, 3))), 'test')
        data.coords = TestCoords()
        self.data = data
        self.px = CoordinateComponent(data, 2)
        self.py = CoordinateComponent(data, 1)
        self.pz = CoordinateComponent(data, 0)
        self.wx = CoordinateComponent(data, 2, world=True)
        self.wy = CoordinateComponent(data, 1, world=True)
        self.wz = CoordinateComponent(data, 0, world=True)

    def test_data(self):
        z, y, x = np.mgrid[0:3, 0:3, 0:3]
        np.testing.assert_array_equal(self.px.data, x)
        np.testing.assert_array_equal(self.py.data, y)
        np.testing.assert_array_equal(self.pz.data, z)
        np.testing.assert_array_equal(self.wx.data, x * 1)
        np.testing.assert_array_equal(self.wy.data, y * 2)
        np.testing.assert_array_equal(self.wz.data, z * 3)

    @pytest.mark.parametrize(('view'), VIEWS)
    def test_view(self, view):
        z, y, x = np.mgrid[0:3, 0:3, 0:3]

        np.testing.assert_array_equal(self.px[view], x[view])
        np.testing.assert_array_equal(self.py[view], y[view])
        np.testing.assert_array_equal(self.pz[view], z[view])
        np.testing.assert_array_equal(self.wx[view], x[view] * 1)
        np.testing.assert_array_equal(self.wy[view], y[view] * 2)
        np.testing.assert_array_equal(self.wz[view], z[view] * 3)


def check_binary(result, left, right, op):
    assert isinstance(result, core.subset.InequalitySubsetState)
    assert result.left is left
    assert result.right is right
    assert result.operator is op


def check_link(result, left, right):
    assert isinstance(result, core.component_link.ComponentLink)
    if isinstance(left, ComponentID):
        assert left in result.get_from_ids()
    if isinstance(right, ComponentID):
        assert right in result.get_from_ids()


# componentID overload
COMPARE_OPS = (operator.gt, operator.ge, operator.lt, operator.le)
NUMBER_OPS = (operator.add, operator.mul, operator.truediv, operator.sub)


@pytest.mark.parametrize(('op'), COMPARE_OPS)
def test_inequality_scalar(op):
    cid = ComponentID('test')
    result = op(cid, 3)
    check_binary(result, cid, 3, op)


@pytest.mark.parametrize(('op'), COMPARE_OPS)
def test_inequality_id(op):
    cid = ComponentID('test')
    cid2 = ComponentID('test2')
    result = op(cid, cid2)
    check_binary(result, cid, cid2, op)


@pytest.mark.parametrize(('op'), NUMBER_OPS)
def test_arithmetic_scalar(op):
    cid = ComponentID('test')
    result = op(cid, 3)
    check_link(result, cid, 3)


@pytest.mark.parametrize(('op'), NUMBER_OPS)
def test_arithmetic_scalar_right(op):
    cid = ComponentID('test')
    result = op(3, cid)
    check_link(result, 3, cid)


@pytest.mark.parametrize(('op'), NUMBER_OPS)
def test_arithmetic_cid(op):
    cid = ComponentID('test')
    cid2 = ComponentID('test2')
    result = op(cid, cid2)
    check_link(result, cid, cid2)


def test_pow_scalar():
    cid = ComponentID('test')
    result = cid ** 3
    check_link(result, cid, 3)


@pytest.mark.parametrize(('view'), VIEWS)
def test_view(view):
    comp = Component(np.array([[1, 2, 3], [2, 3, 4]]))
    np.testing.assert_array_equal(comp[view], comp.data[view])


@pytest.mark.parametrize(('view'), VIEWS)
def test_view_derived(view):
    comp = Component(np.array([[1, 2, 3], [2, 3, 4]]))
    d = core.Data()
    cid = d.add_component(comp, 'primary')
    cid2 = ComponentID("derived")
    link = core.ComponentLink([cid], cid2, using=lambda x: x * 3)
    dc = DerivedComponent(d, link)

    np.testing.assert_array_equal(dc[view], comp.data[view] * 3)


@requires_astropy
def test_units():

    # Make sure that units get converted to strings. At the moment if these
    # are set to Astropy units for example, things can go wrong for example
    # when writing out the datasets. Once we settle on a units framework, we
    # can then use that instead of converting units to strings.

    from astropy import units as u

    comp = Component([1, 2, 3], units='m')
    assert comp.units == 'm'
    assert isinstance(comp.units, str)

    comp = Component([1, 2, 3], units=u.m)
    assert comp.units == 'm'
    assert isinstance(comp.units, str)


def test_scalar_derived():
    # Regression test for a bug that caused expressions without components to
    # fail. Note that the data collection is needed here as the bug occurs in
    # the link manager set up by the data collection.
    data = Data(a=[3, 4, 1])
    dc = DataCollection([data])  # noqa
    pc = ParsedCommand('3.5', {})
    link = ParsedComponentLink(ComponentID('b'), pc)
    data.add_component_link(link)
    np.testing.assert_equal(data['b'], [3.5, 3.5, 3.5])


def test_rename_cid_used_in_derived():
    # Unit test to make sure expressions still work when renaming components
    data = Data(a=[3, 4, 1])
    dc = DataCollection([data])  # noqa
    pc = ParsedCommand('1 + {a}', {'a': data.id['a']})
    link = ParsedComponentLink(ComponentID('b'), pc)
    data.add_component_link(link)
    np.testing.assert_equal(data['b'], [4, 5, 2])
    data.id['a'].label = 'x'
    np.testing.assert_equal(data['b'], [4, 5, 2])


@pytest.mark.xfail
def test_update_cid_used_in_derived():
    # Unit test to make sure expressions still work when updating the component
    # id for components
    data = Data(a=[3, 4, 1])
    dc = DataCollection([data])  # noqa
    pc = ParsedCommand('1 + {a}', {'a': data.id['a']})
    link = ParsedComponentLink(ComponentID('b'), pc)
    data.add_component_link(link)
    np.testing.assert_equal(data['b'], [4, 5, 2])
    data.update_id(data.id['a'], ComponentID('x'))
    np.testing.assert_equal(data['b'], [4, 5, 2])



from itertools import product

import pytest
import numpy as np
from numpy.testing import assert_equal, assert_allclose

from glue.tests.helpers import HYPOTHESIS_INSTALLED

from ..array import (view_shape, coerce_numeric, stack_view, unique, broadcast_to,
                     shape_to_string, check_sorted, pretty_number, unbroadcast,
                     iterate_chunks, combine_slices, nanmean, nanmedian, nansum,
                     nanmin, nanmax, format_minimal, compute_statistic, categorical_ndarray,
                     index_lookup, broadcast_arrays_minimal)


@pytest.mark.parametrize(('before', 'ref_after', 'ref_indices'),
                         (([2.2, 5, 4, 4, 2, 8.3, 2.2], [2, 2.2, 4, 5, 8.3], [1, 3, 2, 2, 0, 4, 1]),
                          ([2.2, 5, np.nan, 2, 8.3, 2.2], [2, 2.2, 5, 8.3], [1, 2, -1, 0, 3, 1])))
def test_unique(before, ref_after, ref_indices):
    after, indices = unique(before)
    np.testing.assert_array_equal(after, ref_after)
    np.testing.assert_array_equal(indices, ref_indices)


def test_unique_dtype():

    # Regression test to make sure that when working with strings, we don't
    # get an object array back from the unique function

    array = np.array(['a', 'b', 'c'])
    U, I = unique(array)
    assert U.dtype.kind in 'SU'

    li = ['a', 'b', 'c']
    U, I = unique(li)
    assert U.dtype.kind in 'SU'


def test_shape_to_string():
    assert shape_to_string((1, 4, 3)) == "(1, 4, 3)"


def test_view_shape():
    assert view_shape((10, 10), np.s_[:]) == (10, 10)
    assert view_shape((10, 10, 10), np.s_[:]) == (10, 10, 10)
    assert view_shape((10, 10), np.s_[:, 1]) == (10,)
    assert view_shape((10, 10), np.s_[2:3, 2:3]) == (1, 1)
    assert view_shape((10, 10), None) == (10, 10)
    assert view_shape((10, 10), ([1, 2, 3], [2, 3, 4])) == (3,)


def test_coerce_numeric():

    x = np.array(['1', '2', '3.14', '4'], dtype=str)
    np.testing.assert_array_equal(coerce_numeric(x),
                                  [1, 2, 3.14, 4])

    x = np.array([1, 2, 3])
    assert coerce_numeric(x) is x

    x = np.array([0, 1, 1, 0], dtype=bool)
    np.testing.assert_array_equal(coerce_numeric(x), np.array([0, 1, 1, 0], dtype=np.int))


@pytest.mark.parametrize(('shape', 'views'),
                         [
                             [(5, 5), (np.s_[0:3],)],
                             [(5, 4), (np.s_[0:3],)],
                             [(5, 4), ((3, 2),)],
                             [(5, 4), (np.s_[0:4], np.s_[:, 0:2])],
                             [(5, 4), (np.s_[0:3, 0:2], 'transpose', (0, 0))],
                             [(10, 20), (np.random.random((10, 20)) > 0.1, 3)],
                             [(5, 7), ('transpose', (3, 2))],
])
def test_stack_view(shape, views):
    x = np.random.random(shape)
    exp = x
    for v in views:
        if isinstance(v, str) and v == 'transpose':
            exp = exp.T
        else:
            exp = exp[v]

    actual = x[stack_view(shape, *views)]

    np.testing.assert_array_equal(exp, actual)


@pytest.mark.parametrize(('array', 'is_sorted'),
                         (([1, 3, 4, 3], False), ([1, 2, np.nan, 3], True), ([1, 3, 4, 4.1], True)))
def test_check_sorted(array, is_sorted):
    assert check_sorted(array) is is_sorted


class TestPrettyNumber(object):

    def test_single(self):
        assert pretty_number([1]) == ['1']
        assert pretty_number([0]) == ['0']
        assert pretty_number([-1]) == ['-1']
        assert pretty_number([1.0001]) == ['1']
        assert pretty_number([1.01]) == ['1.01']
        assert pretty_number([1e-5]) == ['1.000e-05']
        assert pretty_number([1e5]) == ['1.000e+05']
        assert pretty_number([3.3]) == ['3.3']
        assert pretty_number([1.]) == ['1']
        assert pretty_number([1.200]) == ['1.2']

    def test_large(self):
        # Regression test or a bug that caused trailing zeros in exponent to
        # be removed.
        assert pretty_number([1e9]) == ['1.000e+09']
        assert pretty_number([2e10]) == ['2.000e+10']
        assert pretty_number([3e11]) == ['3.000e+11']

    def test_list(self):
        assert pretty_number([1, 2, 3.3, 1e5]) == ['1', '2', '3.3',
                                                   '1.000e+05']


def test_unbroadcast():

    x = np.array([1, 2, 3])
    y = broadcast_to(x, (2, 4, 3))

    z = unbroadcast(y)
    assert z.shape == (1, 1, 3)
    np.testing.assert_allclose(z[0, 0], x)


def test_broadcast_arrays_minimal():

    a = np.array([1, 2, 3])
    b = broadcast_to(a, (2, 4, 3))

    c = np.ones((2, 1, 1))
    d = broadcast_to(c, (2, 4, 3))

    e, f = broadcast_arrays_minimal(b, d)
    assert e.shape == (2, 1, 3)
    assert f.shape == (2, 1, 3)


@pytest.mark.parametrize(('chunk_shape', 'n_max'),
                         [(None, 121), (None, 100000), (None, 5), ((3, 1, 2, 4, 1, 2), None)])
def test_iterate_chunks(chunk_shape, n_max):

    array = np.zeros((6, 3, 4, 5, 1, 8))

    for slices in iterate_chunks(array.shape, chunk_shape=chunk_shape, n_max=n_max):

        print(slices)

        # Make sure empty slices aren't returned
        assert array[slices].size > 0

        # Increment all values in slice by 1
        array[slices] += 1

    assert_equal(array, 1)


def test_iterate_chunks_invalid():

    array = np.zeros((6, 3, 4, 5, 1, 8))

    with pytest.raises(ValueError) as exc:
        next(iterate_chunks(array.shape, chunk_shape=(6, 2, 1)))
    assert exc.value.args[0] == 'chunk_shape should have the same length as shape'

    with pytest.raises(ValueError) as exc:
        next(iterate_chunks(array.shape, chunk_shape=(6, 2, 1, 5, 2, 8)))
    assert exc.value.args[0] == 'chunk_shape should fit within shape'


FUNCTIONS = [nanmean, nanmedian, nanmin, nanmax, nansum]
AXIS = [None, 0, 2, 3, (0, 1), (2, 3), (0, 1, 2), (0, 1, 2, 3)]
ARRAY = np.random.random((4, 5, 2, 7))
ARRAY[ARRAY < 0.1] = np.nan


@pytest.mark.parametrize(('function', 'axis'), product(FUNCTIONS, AXIS))
def test_nanfunctions(function, axis):
    name = function.__name__
    np_func = getattr(np, name)
    assert_allclose(function(ARRAY, axis=axis), np_func(ARRAY, axis=axis))


SLICE_CASES = [
    (slice(None), slice(5), 10),
    (slice(None), slice(1, 5), 10),
    (slice(None), slice(1, 5, 2), 10),
    (slice(1, 8), slice(1, 5), 10),
    (slice(1, 8), slice(1, 5, 2), 10),
    (slice(1, 9, 2), slice(2, 5), 10),
    (slice(1, 9, 2), slice(2, 6), 10),
    (slice(1, 20, 2), slice(3, 18, 2), 20),
    (slice(1, 20, 2), slice(4, 18, 2), 20),
    (slice(1, 20, 2), slice(4, 18, 3), 20),
    (slice(1, None), slice(None, None, 2), 2),
    (slice(2), slice(None), 3)]


@pytest.mark.parametrize(('slice1', 'slice2', 'length'), SLICE_CASES)
def test_combine_slices(slice1, slice2, length):

    # Rather than hard-code the expected result, we can directly check that
    # the resulting slice gives the same indices if applied after the first
    # compared to a manual check

    indices1 = list(range(*slice1.indices(length)))
    indices2 = list(range(*slice2.indices(length)))

    expected = [indices1.index(idx) for idx in indices2 if idx in indices1]

    actual = list(range(*combine_slices(slice1, slice2, length).indices(length)))

    assert actual == expected


if HYPOTHESIS_INSTALLED:

    from hypothesis import given, settings
    from hypothesis.strategies import none, integers

    @given(beg1=none() | integers(-100, 100),
           end1=none() | integers(-100, 100),
           stp1=none() | integers(1, 100),
           beg2=none() | integers(-100, 100),
           end2=none() | integers(-100, 100),
           stp2=none() | integers(1, 100),
           length=integers(0, 100))
    @settings(max_examples=10000, derandomize=True)
    def test_combine_slices_hypot(beg1, end1, stp1, beg2, end2, stp2, length):

        slice1 = slice(beg1, end1, stp1)
        slice2 = slice(beg2, end2, stp2)

        # Rather than hard-code the expected result, we can directly check that
        # the resulting slice gives the same indices if applied after the first
        # compared to a manual check

        indices1 = list(range(*slice1.indices(length)))
        indices2 = list(range(*slice2.indices(length)))

        expected = [indices1.index(idx) for idx in indices2 if idx in indices1]

        actual = list(range(*combine_slices(slice1, slice2, length).indices(length)))

        assert actual == expected


def test_format_minimal():

    # TODO: in future could consider detecting integer cases
    fmt, strings = format_minimal([133, 1444, 3300])
    assert fmt == "{:.1f}"
    assert strings == ['133.0', '1444.0', '3300.0']

    # TODO: in future could consider detecting if all intervals are integers
    fmt, strings = format_minimal([133., 1444., 3300.])
    assert fmt == "{:.1f}"
    assert strings == ['133.0', '1444.0', '3300.0']

    fmt, strings = format_minimal([3, 4.3, 4.4, 5])
    assert fmt == "{:.1f}"
    assert strings == ['3.0', '4.3', '4.4', '5.0']

    fmt, strings = format_minimal([3, 4.325, 4.326, 5])
    assert fmt == "{:.3f}"
    assert strings == ['3.000', '4.325', '4.326', '5.000']

    fmt, strings = format_minimal([-3, 0., 0.993e-4, 5])
    assert fmt == "{:.4f}"
    assert strings == ['-3.0000', '0.0000', '0.0001', '5.0000']

    fmt, strings = format_minimal([-3, 0., 0.993e-8, 5])
    assert fmt == "{:.1e}"
    assert strings == ['-3.0e+00', '0.0e+00', '9.9e-09', '5.0e+00']


CASES = [('mean', [-3, 1, 6], dict(), 4 / 3),
         ('mean', [-3, 1, 6], dict(mask=[1, 0, 1]), 1.5),
         ('mean', [-3, 1, 6], dict(positive=True), 3.5),
         ('mean', [-3, 1, np.nan], dict(finite=True), -1),
         ('mean', [-3, 1, np.nan], dict(finite=False), np.nan),
         ('median', [-3, 1, 6], dict(), 1),
         ('median', [-3, 1, 6, 7], dict(mask=[1, 1, 0, 1]), 1),
         ('median', [-3, 1, 6, 7], dict(positive=True), 6),
         ('median', [-3, 1, np.nan, 6], dict(finite=True), 1),
         ('median', [-3, 1, np.nan, 6], dict(finite=False), np.nan),
         ('maximum', [-3, 1, 6], dict(), 6),
         ('minimum', [-3, 1, 6], dict(), -3),
         ('sum', [-3, 1, 6], dict(), 4),
         ('percentile', [-3, 1, 6], dict(percentile=0), -3),
         ('percentile', [-3, 1, 6], dict(percentile=50), 1),
         ('percentile', [-3, 1, 6], dict(percentile=100), 6),
         ('mean', [-3, 1, 6], dict(axis=()), [-3, 1, 6]),
         ('maximum', [-3, 1, 6], dict(axis=()), [-3, 1, 6]),
         ('mean', [-3, 1, 6], dict(axis=0), 4 / 3),
         ('mean', [-3, 1, 6], dict(axis=(0,)), 4 / 3),
         ('mean', [[-1, 2], [3, 4]], dict(), 2),
         ('mean', [[-1, 2], [3, 4]], dict(axis=(0, 1)), 2),
         ('mean', [[-1, 2], [3, 4]], dict(axis=0), [1, 3]),
         ('mean', [[-1, 2], [3, 4]], dict(axis=1), [0.5, 3.5]),
         ('mean', [[-1, 2], [3, 4]], dict(axis=1, positive=True), [2, 3.5]),
         ('mean', [[np.nan, 2], [3, 4]], dict(axis=1, finite=True), [2, 3.5]),
         ('mean', [[np.nan, 2], [3, 4]], dict(axis=1, finite=False), [np.nan, 3.5]),
         ('mean', [], dict(), np.nan),
         ('mean', [-3, 1, 6], dict(mask=[0, 0, 0]), np.nan),
         ]


@pytest.mark.parametrize(('statistic', 'data', 'kwargs', 'result'), CASES)
def test_compute_statistic(statistic, data, kwargs, result):
    assert_allclose(compute_statistic(statistic, data, **kwargs), result)


def test_index_lookup():

    assert_equal(index_lookup(['a', 'b', 't', 'f', 'a', 'b'], ['a', 'b', 'f']),
                 [0, 1, np.nan, 2, 0, 1])


def test_categorical_ndarray():

    array = categorical_ndarray(['a', 'b', 'c', 'b', 'b', 'a'])
    assert_equal(array, ['a', 'b', 'c', 'b', 'b', 'a'])
    assert_equal(array.codes, [0, 1, 2, 1, 1, 0])
    assert_equal(array.categories, ['a', 'b', 'c'])

    array = categorical_ndarray(['a', 'b', 'c', 'b', 'b', 'a'], categories=['b', 'a'])
    assert_equal(array, ['a', 'b', 'c', 'b', 'b', 'a'])
    assert_equal(array.codes, [1, 0, np.nan, 0, 0, 1])
    assert_equal(array.categories, ['b', 'a'])

    np.random.seed(12345)
    array.jitter(method='uniform')
    assert_allclose(array.codes, [1.42961609, -0.18362445, np.nan,
                                  -0.29543972, 0.06772503, 1.0955447])

    array.jitter(method=None)
    assert_equal(array.codes, [1, 0, np.nan, 0, 0, 1])


def test_categorical_ndarray_slicing():

    array = categorical_ndarray(['a', 'b', 'c', 'b', 'b', 'a'])
    assert_equal(array, ['a', 'b', 'c', 'b', 'b', 'a'])
    assert_equal(array.codes, [0, 1, 2, 1, 1, 0])
    assert_equal(array.categories, ['a', 'b', 'c'])

    subset = array[1:5]
    assert_equal(subset, ['b', 'c', 'b', 'b'])
    assert_equal(subset.codes, [1, 2, 1, 1])
    assert_equal(subset.categories, ['a', 'b', 'c'])



# New contrast/bias mode that operates on viewers with state objects

from echo import delay_callback
from glue.config import viewer_tool
from glue.viewers.matplotlib.toolbar_mode import ToolbarModeBase


@viewer_tool
class ContrastBiasMode(ToolbarModeBase):
    """
    Uses right mouse button drags to set bias and contrast, DS9-style.    The
    horizontal position of the mouse sets the bias, the vertical position sets
    the contrast.
    """

    icon = 'glue_contrast'
    tool_id = 'image:contrast_bias'
    action_text = 'Contrast/Bias'
    tool_tip = 'Adjust the bias/contrast'
    status_tip = ('CLICK and DRAG on image from left to right to adjust '
                  'bias and up and down to adjust contrast')

    def move(self, event):
        """
        Update bias and contrast on Right Mouse button drag.
        """

        if event.button not in (1, 3):
            return

        x, y = self.viewer.axes.transAxes.inverted().transform((event.x, event.y))
        state = self.viewer.selected_layer.state

        with delay_callback(state, 'bias', 'contrast'):
            state.bias = -(x * 2 - 1.5)
            state.contrast = 10. ** (y * 2 - 1)

        super(ContrastBiasMode, self).move(event)



from qtpy.QtCore import QTimer

from glue.core.message import ComputationStartedMessage
from glue.viewers.common.qt.data_viewer import DataViewer
from glue.viewers.matplotlib.qt.widget import MplWidget
from glue.viewers.matplotlib.mpl_axes import init_mpl
from glue.utils import defer_draw, decorate_all_methods
from glue.viewers.matplotlib.state import MatplotlibDataViewerState
from glue.viewers.matplotlib.viewer import MatplotlibViewerMixin

# The following import is required to register the viewer tools
from glue.viewers.matplotlib.qt import toolbar  # noqa

__all__ = ['MatplotlibDataViewer']


@decorate_all_methods(defer_draw)
class MatplotlibDataViewer(MatplotlibViewerMixin, DataViewer):

    _state_cls = MatplotlibDataViewerState

    tools = ['mpl:home', 'mpl:pan', 'mpl:zoom']
    subtools = {'save': ['mpl:save']}

    def __init__(self, session, parent=None, wcs=None, state=None, projection=None):

        super(MatplotlibDataViewer, self).__init__(session, parent=parent, state=state)

        # Use MplWidget to set up a Matplotlib canvas inside the Qt window
        self.mpl_widget = MplWidget()
        self.setCentralWidget(self.mpl_widget)

        # TODO: shouldn't have to do this
        self.central_widget = self.mpl_widget

        self.figure, self.axes = init_mpl(self.mpl_widget.canvas.fig, wcs=wcs, projection=projection)

        MatplotlibViewerMixin.setup_callbacks(self)

        self.central_widget.resize(600, 400)
        self.resize(self.central_widget.size())

        self._monitor_computation = QTimer()
        self._monitor_computation.setInterval(500)
        self._monitor_computation.timeout.connect(self._update_computation)

    def _update_computation(self, message=None):

        # If we get a ComputationStartedMessage and the timer isn't currently
        # active, then we start the timer but we then return straight away.
        # This is to avoid showing the 'Computing' message straight away in the
        # case of reasonably fast operations.
        if isinstance(message, ComputationStartedMessage):
            if not self._monitor_computation.isActive():
                self._monitor_computation.start()
            return

        for layer_artist in self.layers:
            if layer_artist.is_computing:
                self.loading_rectangle.set_visible(True)
                text = self.loading_text.get_text()
                if text.count('.') > 2:
                    text = 'Computing'
                else:
                    text += '.'
                self.loading_text.set_text(text)
                self.loading_text.set_visible(True)
                self.redraw()
                return

        self.loading_rectangle.set_visible(False)
        self.loading_text.set_visible(False)
        self.redraw()

        # If we get here, the computation has stopped so we can stop the timer
        self._monitor_computation.stop()



from glue.viewers.matplotlib.mpl_axes import init_mpl

from matplotlib.figure import Figure
import numpy as np


class ProjectionMplTransform(object):
    def __init__(self, projection, x_lim, y_lim, x_scale, y_scale):
        self._state = {'projection': projection, 'x_lim': x_lim, 'y_lim': y_lim,
                       'x_scale': x_scale, 'y_scale': y_scale}
        _, axes = init_mpl(Figure(), projection=self._state['projection'])
        axes.set_xscale(self._state['x_scale'])
        axes.set_yscale(self._state['y_scale'])
        if self._state['projection'] not in ['aitoff', 'hammer', 'lambert', 'mollweide']:
            axes.set_xlim(self._state['x_lim'])
            axes.set_ylim(self._state['y_lim'])
        self._transform = (axes.transData + axes.transAxes.inverted()).frozen()

    def __call__(self, x, y):
        points = np.hstack((x.reshape(-1, 1), y.reshape(-1, 1)))
        res = self._transform.transform(points)
        out = np.hsplit(res, 2)
        return out[0].reshape(x.shape), out[1].reshape(y.shape)

    def __gluestate__(self, context):
        return dict(state=context.do(self._state))

    @classmethod
    def __setgluestate__(cls, rec, context):
        state = context.object(rec['state'])
        return cls(state['projection'], state['x_lim'], state['y_lim'],
                   state['x_scale'], state['y_scale'])



from ..config import qt_client, link_function, data_factory
from glue.tests.helpers import requires_qt


@requires_qt
def test_default_clients():

    from glue.viewers.image.qt import ImageViewer
    from glue.viewers.scatter.qt import ScatterViewer
    from glue.viewers.histogram.qt import HistogramViewer

    assert ImageViewer in qt_client
    assert ScatterViewer in qt_client
    assert HistogramViewer in qt_client


def test_add_client():
    @qt_client
    class TestClient(object):
        pass

    assert TestClient in qt_client


def test_add_link_default():
    @link_function(info='maps x to y', output_labels=['y'])
    def foo(x):
        return 3
    val = (foo, 'maps x to y', ['y'], 'General')
    assert val in link_function


def test_add_data_factory():
    @data_factory('XYZ file', "*txt")
    def foo(x):
        pass
    assert (foo, 'XYZ file', '*txt', 0, False) in data_factory



from qtpy import QtCore
from qtpy.QtCore import Qt

__all__ = ['PythonListModel']


class PythonListModel(QtCore.QAbstractListModel):

    """
    A Qt Model that wraps a python list, and exposes a list-like interface

    This can be connected directly to multiple QListViews, which will
    stay in sync with the state of the container.
    """

    def __init__(self, items, parent=None):
        """
        Create a new model

        Parameters
        ----------
        items : list
            The initial list to wrap
        parent : QObject
            The model parent
        """
        super(PythonListModel, self).__init__(parent)
        self.items = items

    def rowCount(self, parent=None):
        """Number of rows"""
        return len(self.items)

    def headerData(self, section, orientation, role):
        """Column labels"""
        if role != Qt.DisplayRole:
            return None
        return "%i" % section

    def row_label(self, row):
        """ The textual label for the row"""
        return str(self.items[row])

    def data(self, index, role):
        """Retrieve data at each index"""
        if not index.isValid():
            return None
        if role == Qt.DisplayRole or role == Qt.EditRole:
            return self.row_label(index.row())
        if role == Qt.UserRole:
            return self.items[index.row()]

    def setData(self, index, value, role):
        """
        Update the data in-place

        Parameters
        ----------
        index : QModelIndex
            The location of the change
        value : object
            The new value
        role : QEditRole
            Which aspect of the model to update
        """
        if not index.isValid():
            return False

        if role == Qt.UserRole:
            row = index.row()
            self.items[row] = value
            self.dataChanged.emit(index, index)
            return True

        return super(PythonListModel, self).setDdata(index, value, role)

    def removeRow(self, row, parent=None):
        """
        Remove a row from the table

        Parameters
        ----------
        row : int
            Row to remove

        Returns
        -------
        successful : bool
        """
        if row < 0 or row >= len(self.items):
            return False

        self.beginRemoveRows(QtCore.QModelIndex(), row, row)
        self._remove_row(row)
        self.endRemoveRows()
        return True

    def pop(self, row=None):
        """
        Remove and return an item (default last item)

        Parameters
        ----------
        row : int (optional)
            Which row to remove. Default=last

        Returns
        --------
        popped : object
        """
        if row is None:
            row = len(self) - 1
        result = self[row]
        self.removeRow(row)
        return result

    def _remove_row(self, row):
        # actually remove data. Subclasses can override this as needed
        self.items.pop(row)

    def __getitem__(self, row):
        return self.items[row]

    def __setitem__(self, row, value):
        index = self.index(row)
        self.setData(index, value, role=Qt.UserRole)

    def __len__(self):
        return len(self.items)

    def insert(self, row, value):
        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        self.items.insert(row, value)
        self.endInsertRows()
        self.rowsInserted.emit(self.index(row), row, row)

    def append(self, value):
        row = len(self)
        self.insert(row, value)

    def extend(self, values):
        for v in values:
            self.append(v)

    def set_list(self, values):
        """
        Set the model to a new list
        """
        self.beginResetModel()
        self.items = values
        self.endResetModel()



import os
import sys
import warnings

import pytest

try:
    from qtpy import PYSIDE2
except Exception:
    PYSIDE2 = False

from glue.config import CFG_DIR as CFG_DIR_ORIG

try:
    import objgraph
except ImportError:
    OBJGRAPH_INSTALLED = False
else:
    OBJGRAPH_INSTALLED = True


STDERR_ORIGINAL = sys.stderr

ON_APPVEYOR = os.environ.get('APPVEYOR', 'False') == 'True'


def pytest_runtest_teardown(item, nextitem):
    sys.stderr = STDERR_ORIGINAL
    global start_dir
    os.chdir(start_dir)


def pytest_addoption(parser):
    parser.addoption("--no-optional-skip", action="store_true", default=False,
                     help="don't skip any tests with optional dependencies")


start_dir = None


def pytest_configure(config):

    global start_dir
    start_dir = os.path.abspath('.')

    os.environ['GLUE_TESTING'] = 'True'

    if config.getoption('no_optional_skip'):
        from glue.tests import helpers
        for attr in helpers.__dict__:
            if attr.startswith('requires_'):
                # The following line replaces the decorators with a function
                # that does noting, effectively disabling it.
                setattr(helpers, attr, lambda f: f)

    # Make sure we don't affect the real glue config dir
    import tempfile
    from glue import config
    config.CFG_DIR = tempfile.mkdtemp()

    # Start up QApplication, if the Qt code is present
    try:
        from glue.utils.qt import get_qapp
    except Exception:
        # Note that we catch any exception, not just ImportError, because
        # QtPy can raise a PythonQtError.
        pass
    else:
        get_qapp()

    # Force loading of plugins
    from glue.main import load_plugins
    load_plugins()


def pytest_report_header(config):
    from glue import __version__
    glue_version = "%20s:\t%s" % ("glue", __version__)
    from glue._deps import get_status
    return os.linesep + glue_version + os.linesep + os.linesep + get_status()


def pytest_unconfigure(config):

    os.environ.pop('GLUE_TESTING')

    # Reset configuration directory to original one
    from glue import config
    config.CFG_DIR = CFG_DIR_ORIG

    # Remove reference to QApplication to prevent segmentation fault on PySide
    try:
        from glue.utils.qt import app
        app.qapp = None
    except Exception:  # for when we run the tests without the qt directories
        # Note that we catch any exception, not just ImportError, because
        # QtPy can raise a PythonQtError.
        pass

    if OBJGRAPH_INSTALLED and not ON_APPVEYOR:

        # Make sure there are no lingering references to GlueApplication
        obj = objgraph.by_type('GlueApplication')
        if len(obj) > 0:
            objgraph.show_backrefs(objgraph.by_type('GlueApplication'))
            warnings.warn('There are {0} remaining references to GlueApplication'.format(len(obj)))

        # Uncomment when checking for memory leaks
        # objgraph.show_most_common_types(limit=100)


# With PySide2, tests can fail in a non-deterministic way with the following
# error:
#
#   AttributeError: 'PySide2.QtGui.QStandardItem' object has no attribute 'connect'
#
# Until this can be properly debugged and fixed, we xfail any test that fails
# with this exception

if PYSIDE2:

    def pytest_runtest_call(__multicall__):
        try:
            __multicall__.execute()
        except AttributeError as exc:
            if 'PySide2.QtGui.QStandardItem' in str(exc):
                pytest.xfail()



# pylint: disable=I0011,W0613,W0201,W0212,E1101,E1103

from collections import namedtuple

import pytest
import numpy as np
from numpy.testing import assert_allclose
import matplotlib.pyplot as plt
from unittest.mock import MagicMock
from matplotlib.figure import Figure
from numpy.testing import assert_almost_equal

from .. import roi as r
from ..component import CategoricalComponent
from ..roi import (RectangularROI, UndefinedROI, CircularROI, PolygonalROI, CategoricalROI,
                   MplCircularROI, MplRectangularROI, MplPolygonalROI, MplPickROI, PointROI,
                   XRangeROI, MplXRangeROI, YRangeROI, MplYRangeROI, RangeROI, Projected3dROI,
                   EllipticalROI)
from ..state import GlueSerializer, GlueUnSerializer


FIG = Figure()
AXES = FIG.add_subplot(111)


def roundtrip_roi(roi):
    gs = GlueSerializer(roi)
    out_str = gs.dumps()
    obj = GlueUnSerializer.loads(out_str)
    return obj.object('__main__')


class TestPoint(object):

    def setup_method(self, method):
        self.roi = PointROI(1, 2)

    def test_contains(self):
        assert not self.roi.contains(3, 3)
        assert not self.roi.contains(1, 2)

    def test_move_to(self):
        self.roi.move_to(4, 5)
        assert self.roi.x == 4
        assert self.roi.y == 5

    def test_defined(self):
        assert self.roi.defined()

    def test_not_defined(self):
        self.roi.reset()
        assert not self.roi.defined()

    def test_center(self):
        assert self.roi.center() == (1, 2)


class TestRectangle(object):

    def setup_method(self, method):
        self.roi = RectangularROI()

    def test_empty_roi_contains_raises(self):
        with pytest.raises(UndefinedROI):
            self.roi.contains(1, 2)

    def test_scalar_contains(self):
        self.roi.update_limits(0, 0, 10, 10)
        assert self.roi.contains(5, 5)
        assert not self.roi.contains(11, 11)

    def test_reset(self):
        assert not self.roi.defined()
        self.roi.update_limits(0, 0, 10, 10)
        assert self.roi.defined()
        self.roi.reset()
        assert not self.roi.defined()
        with pytest.raises(UndefinedROI):
            self.roi.contains(5, 5)

    def test_empty_to_polygon(self):
        x, y = self.roi.to_polygon()
        assert x == []
        assert y == []

    def test_to_polygon(self):
        self.roi.update_limits(0, 0, 10, 10)
        x, y = self.roi.to_polygon()
        poly = PolygonalROI(vx=x, vy=y)
        assert poly.contains(5, 5)

    def test_ndarray(self):
        self.roi.update_limits(0, 0, 10, 10)

        x = np.array([5, 6, 2, 11])
        y = np.array([5, 11, 2, 11])
        result = self.roi.contains(x, y)
        assert result[0]
        assert not result[1]
        assert result[2]
        assert not result[3]

    def test_corner(self):
        self.roi.update_limits(6, 7, 10, 10)
        assert self.roi.corner() == (6, 7)

    def test_width(self):
        self.roi.update_limits(2, 2, 10, 12)
        assert self.roi.width() == 8

    def test_height(self):
        self.roi.update_limits(2, 2, 10, 12)
        assert self.roi.height() == 10

    def test_multidim_ndarray(self):
        self.roi.update_limits(0, 0, 10, 10)
        x = np.array([1, 2, 3, 4]).reshape(2, 2)
        y = np.array([1, 2, 3, 4]).reshape(2, 2)
        assert self.roi.contains(x, y).all()
        assert not self.roi.contains(x + 10, y).any()
        assert self.roi.contains(x, y).shape == x.shape

    def test_str_undefined(self):
        """ str method should not crash """
        assert type(str(self.roi)) == str

    def test_str_defined(self):
        """ str method should not crash """
        self.roi.update_limits(1, 2, 3, 4)
        assert type(str(self.roi)) == str

    def test_serialization(self):
        self.roi.update_limits(1, 2, 3, 4)
        new_roi = roundtrip_roi(self.roi)
        assert_almost_equal(new_roi.xmin, 1)
        assert_almost_equal(new_roi.ymin, 2)
        assert_almost_equal(new_roi.xmax, 3)
        assert_almost_equal(new_roi.ymax, 4)


class TestRange(object):

    def test_wrong_orientation(self):
        with pytest.raises(ValueError):
            RangeROI(orientation='a')


class TestXRange(object):

    def test_undefined_on_init(self):
        assert not XRangeROI().defined()

    def test_str(self):
        roi = XRangeROI()
        assert str(roi) == "Undefined XRangeROI"

        roi.set_range(1, 2)
        assert str(roi) == "1.000 < x < 2.000"

    def test_reset(self):
        roi = XRangeROI()
        roi.set_range(1, 2)
        assert roi.defined()
        roi.reset()
        assert not roi.defined()

    def test_contains(self):
        roi = XRangeROI()
        roi.set_range(1, 3)
        x = np.array([0, 1, 2, 3])
        y = np.array([-np.inf, 100, 200, 0])
        np.testing.assert_array_equal(roi.contains(x, y),
                                      [False, False, True, False])

    def test_contains_undefined(self):
        roi = XRangeROI()
        with pytest.raises(UndefinedROI):
            roi.contains(1, 2)

    def test_to_polygon(self):
        roi = XRangeROI()
        assert roi.to_polygon() == ([], [])
        roi.set_range(1, 2)
        x, y = roi.to_polygon()
        np.testing.assert_array_equal(x, [1, 2, 2, 1, 1])
        np.testing.assert_array_equal(y,
                                      [-1e100, -1e100, 1e100, 1e100, -1e100])

    def test_serialization(self):
        roi = XRangeROI()
        roi.set_range(7, 8)
        new_roi = roundtrip_roi(roi)
        assert_almost_equal(new_roi.min, 7)
        assert_almost_equal(new_roi.max, 8)
        assert new_roi.ori == 'x'


class TestYRange(object):
    def test_undefined_on_init(self):
        assert not YRangeROI().defined()

    def test_str(self):
        roi = YRangeROI()
        assert str(roi) == "Undefined YRangeROI"

        roi.set_range(1, 2)
        assert str(roi) == "1.000 < y < 2.000"

    def test_reset(self):
        roi = YRangeROI()
        roi.set_range(1, 2)
        assert roi.defined()
        roi.reset()
        assert not roi.defined()

    def test_contains(self):
        roi = YRangeROI()
        roi.set_range(1, 3)
        y = np.array([0, 1, 2, 3])
        x = np.array([-np.inf, 100, 200, 0])
        np.testing.assert_array_equal(roi.contains(x, y),
                                      [False, False, True, False])

    def test_contains_undefined(self):
        roi = YRangeROI()
        with pytest.raises(UndefinedROI):
            roi.contains(1, 2)

    def test_to_polygon(self):
        roi = YRangeROI()
        assert roi.to_polygon() == ([], [])
        roi.set_range(1, 2)
        x, y = roi.to_polygon()
        np.testing.assert_array_equal(y, [1, 2, 2, 1, 1])
        np.testing.assert_array_equal(x,
                                      [-1e100, -1e100, 1e100, 1e100, -1e100])

    def test_serialization(self):
        roi = YRangeROI()
        roi.set_range(4, 5)
        new_roi = roundtrip_roi(roi)
        assert_almost_equal(new_roi.min, 4)
        assert_almost_equal(new_roi.max, 5)
        assert new_roi.ori == 'y'


class TestCircle(object):
    def setup_method(self, method):
        self.roi = CircularROI()

    def test_undefined_on_creation(self):
        assert not self.roi.defined()

    def test_contains_on_undefined_contains_raises(self):
        with pytest.raises(UndefinedROI):
            self.roi.contains(1, 1)

    def test_set_center(self):
        self.roi.set_center(0, 0)
        self.roi.set_radius(1)
        assert self.roi.contains(0, 0)
        assert not self.roi.contains(2, 2)
        self.roi.set_center(2, 2)
        assert not self.roi.contains(0, 0)
        assert self.roi.contains(2, 2)

    def test_set_radius(self):
        self.roi.set_center(0, 0)
        self.roi.set_radius(1)
        assert not self.roi.contains(1.5, 0)
        self.roi.set_radius(5)
        assert self.roi.contains(1.5, 0)

    def test_contains_many(self):
        x = [0, 0, 0, 0, 0]
        y = [0, 0, 0, 0, 0]
        self.roi.set_center(0, 0)
        self.roi.set_radius(1)
        assert all(self.roi.contains(x, y))
        assert all(self.roi.contains(np.asarray(x), np.asarray(y)))
        assert not any(self.roi.contains(np.asarray(x) + 10, y))

    def test_poly(self):
        self.roi.set_center(0, 0)
        self.roi.set_radius(1)
        x, y = self.roi.to_polygon()
        poly = PolygonalROI(vx=x, vy=y)
        assert poly.contains(0, 0)
        assert not poly.contains(10, 0)

    def test_poly_undefined(self):
        x, y = self.roi.to_polygon()
        assert x == []
        assert y == []

    def test_reset(self):
        assert not self.roi.defined()
        self.roi.set_center(0, 0)
        assert not self.roi.defined()
        self.roi.set_radius(2)
        assert self.roi.defined()
        self.roi.reset()
        assert not self.roi.defined()

    def test_multidim(self):
        self.roi.set_center(0, 0)
        self.roi.set_radius(1)
        x = np.array([.1, .2, .3, .4]).reshape(2, 2)
        y = np.array([-.1, -.2, -.3, -.4]).reshape(2, 2)
        assert self.roi.contains(x, y).all()
        assert not self.roi.contains(x + 1, y).any()
        assert self.roi.contains(x, y).shape == (2, 2)

    def test_serialization(self):
        self.roi.set_center(3, 4)
        self.roi.set_radius(5)
        new_roi = roundtrip_roi(self.roi)
        assert_almost_equal(new_roi.xc, 3)
        assert_almost_equal(new_roi.yc, 4)
        assert_almost_equal(new_roi.radius, 5)


class TestEllipse(object):

    def setup_method(self, method):
        self.roi_empty = EllipticalROI()
        self.roi = EllipticalROI(1, 2, 3, 4)

    def test_undefined_on_creation(self):
        assert not self.roi_empty.defined()
        assert self.roi.defined()

    def test_contains_on_undefined_contains_raises(self):
        with pytest.raises(UndefinedROI):
            self.roi_empty.contains(1, 1)
        self.roi.contains(1, 1)

    def test_set_center(self):
        assert self.roi.contains(0, 0)
        assert not self.roi.contains(12, 12)
        self.roi.xc = 11
        self.roi.yc = 12
        assert not self.roi.contains(0, 0)
        assert self.roi.contains(12, 12)

    def test_set_radius(self):
        assert self.roi.contains(0, 0)
        assert not self.roi.contains(12, 12)
        self.roi.radius_y = 100
        assert self.roi.contains(0, 0)
        assert not self.roi.contains(12, 12)
        self.roi.radius_x = 100
        assert self.roi.contains(0, 0)
        assert self.roi.contains(12, 12)

    def test_contains_many(self):
        x = [0, 0, 0, 0, 0]
        y = [0, 0, 0, 0, 0]
        assert all(self.roi.contains(x, y))
        assert all(self.roi.contains(np.asarray(x), np.asarray(y)))
        assert not any(self.roi.contains(np.asarray(x) + 10, y))

    def test_poly(self):
        x, y = self.roi.to_polygon()
        poly = PolygonalROI(vx=x, vy=y)
        assert poly.contains(0, 0)
        assert not poly.contains(10, 0)

    def test_poly_undefined(self):
        x, y = self.roi_empty.to_polygon()
        assert x == []
        assert y == []

    def test_reset(self):
        assert not self.roi_empty.defined()
        self.roi_empty.xc = 1
        assert not self.roi_empty.defined()
        self.roi_empty.yc = 2
        assert not self.roi_empty.defined()
        self.roi_empty.radius_x = 3
        assert not self.roi_empty.defined()
        self.roi_empty.radius_y = 4
        assert self.roi_empty.defined()
        self.roi_empty.reset()
        assert not self.roi_empty.defined()

    def test_multidim(self):
        x = np.array([.1, .2, .3, .4]).reshape(2, 2)
        y = np.array([-.1, -.2, -.3, -.4]).reshape(2, 2)
        assert self.roi.contains(x, y).all()
        assert not self.roi.contains(x + 10, y).any()
        assert self.roi.contains(x, y).shape == (2, 2)

    def test_serialization(self):
        new_roi = roundtrip_roi(self.roi)
        assert_almost_equal(new_roi.xc, 1)
        assert_almost_equal(new_roi.yc, 2)
        assert_almost_equal(new_roi.radius_x, 3)
        assert_almost_equal(new_roi.radius_y, 4)


class TestPolygon(object):
    def setup_method(self, method):
        self.roi = PolygonalROI()

    def define_as_square(self):
        self.roi.reset()
        assert not self.roi.defined()
        self.roi.add_point(0, 0)
        self.roi.add_point(0, 1)
        self.roi.add_point(1, 1)
        self.roi.add_point(1, 0)
        assert self.roi.defined()

    def test_contains_on_empty_raises(self):
        with pytest.raises(UndefinedROI):
            self.roi.contains(1, 2)

    def test_remove_empty(self):
        self.roi.remove_point(1, 0)

    def test_replace(self):
        self.define_as_square()
        assert self.roi.contains(.9, .02)
        self.roi.replace_last_point(0, 0)
        assert not self.roi.contains(.9, .01)

    def test_remove_successful(self):
        self.define_as_square()
        assert self.roi.contains(.9, .02)
        self.roi.remove_point(1, 0)
        assert not self.roi.contains(.9, .01)

    def test_nan(self):
        self.define_as_square()
        assert not self.roi.contains(np.nan, .5)

    def test_remove_unsuccessful(self):
        self.define_as_square()
        assert self.roi.contains(.9, .02)
        self.roi.remove_point(1.5, 0, thresh=.49)
        assert self.roi.contains(.9, .01)

    def test_to_poly(self):
        self.define_as_square()
        x, y = self.roi.to_polygon()
        assert x == [0, 0, 1, 1]
        assert y == [0, 1, 1, 0]

    def test_to_poly_undefined(self):
        assert self.roi.to_polygon() == ([], [])

    def test_rect(self):
        self.roi.reset()
        self.roi.add_point(4.95164474584, 0.136922625654)
        self.roi.add_point(6.08256580223, 0.136922625654)
        self.roi.add_point(6.08256580223, 0.423771697842)
        self.roi.add_point(4.95164474584, 0.423771697842)
        self.roi.add_point(4.95164474584, 0.136922625654)
        x = np.array([4.4, 3.18, 4.49, 4.49])
        y = np.array([.2, .2, .2, .2])
        assert not self.roi.contains(4.4000001, 0.2)
        assert not self.roi.contains(3.1800001, 0.2)
        assert not self.roi.contains(4.4899998, 0.2)
        assert not self.roi.contains(x, y).any()
        x.shape = (2, 2)
        y.shape = (2, 2)
        assert not self.roi.contains(x, y).any()
        assert self.roi.contains(x, y).shape == (2, 2)

    def test_empty(self):
        assert not self.roi.defined()
        with pytest.raises(UndefinedROI):
            self.roi.contains(0, 0)

    def test_contains_scalar(self):
        self.define_as_square()
        assert self.roi.contains(.5, .5)
        assert not self.roi.contains(1.5, 1.5)

    def test_contains_list(self):
        self.define_as_square()
        assert self.roi.contains([.5, .4], [.5, .4]).all()
        assert not self.roi.contains([1.5, 1.5], [0, 0]).any()

    def test_contains_numpy(self):
        self.define_as_square()
        x = np.array([.4, .5, .4])
        y = np.array([.4, .5, .4])
        assert self.roi.contains(x, y).all()
        assert not self.roi.contains(x + 5, y).any()

    def test_str(self):
        """ __str__ returns a string """
        assert type(str(self.roi)) == str

    def test_serialization(self):
        self.define_as_square()
        new_roi = roundtrip_roi(self.roi)
        assert_almost_equal(new_roi.vx, np.array([0, 0, 1, 1]))
        assert_almost_equal(new_roi.vy, np.array([0, 1, 1, 0]))


class TestProjected3dROI(object):
    # matrix that converts xyzw to yxzw
    xyzw2yxzw = np.array([[0, 1, 0, 0], [0, 0, 1, 0], [1, 0, 0, 0], [0, 0, 0, 1]])
    x = [1, 2, 3]
    y = [2, 3, 4]
    z = [5, 6, 7]
    # repeat the arrays, 'rolled over' by 1
    x_nd = [[1, 3], [2, 1], [3, 2]]
    y_nd = [[2, 3], [3, 2], [4, 3]]
    z_nd = [[5, 7], [6, 5], [7, 6]]

    def test_contains2d(self):
        roi_2d = PolygonalROI(vx=[0.5, 2.5, 2.5, 0.5], vy=[1, 1, 3.5, 3.5])
        roi = Projected3dROI(roi_2d=roi_2d, projection_matrix=np.eye(4))
        assert roi.contains(self.x, self.y).tolist() == [True, True, False]

    def test_contains3d(self):
        roi_2d = PolygonalROI(vx=[1.5, 3.5, 3.5, 1.5], vy=[4, 4, 6.5, 6.5])
        roi = Projected3dROI(roi_2d=roi_2d, projection_matrix=self.xyzw2yxzw)
        assert roi.contains3d(self.x, self.y, self.z).tolist() == [True, True, False]
        assert roi.contains3d(self.x_nd, self.y_nd, self.z_nd).tolist() == [[True, False], [True, True], [False, True]]

    def test_forward(self):
        # testing the calls that should be forwarded to roi_2d
        roi_2d = PolygonalROI(vx=[0.5, 2.5, 2.5, 0.5], vy=[1, 1, 3.5, 3.5])
        roi = Projected3dROI(roi_2d=roi_2d, projection_matrix=np.eye(4))

        assert roi.contains(self.x, self.y).tolist() == roi_2d.contains(self.x, self.y).tolist()
        assert roi.to_polygon() == roi_2d.to_polygon()
        assert roi.defined() == roi_2d.defined()


class TestCategorical(object):

    def test_empty(self):

        roi = CategoricalROI()
        assert not roi.defined()

    def test_defined(self):

        nroi = CategoricalROI(categories=['a', 'b', 'c'])
        assert nroi.defined()
        nroi.reset()
        assert not nroi.defined()

    def test_loads_from_components(self):

        roi = CategoricalROI()
        comp = CategoricalComponent(np.array(['a', 'a', 'b']))
        roi.update_categories(comp)

        np.testing.assert_array_equal(roi.categories,
                                      np.array(['a', 'b']))

        roi = CategoricalROI(categories=comp)
        np.testing.assert_array_equal(roi.categories,
                                      np.array(['a', 'b']))

    def test_applies_components(self):

        roi = CategoricalROI()
        comp = CategoricalComponent(np.array(['a', 'b', 'c']))
        roi.update_categories(CategoricalComponent(np.array(['a', 'b'])))
        contained = roi.contains(comp, None)
        np.testing.assert_array_equal(contained,
                                      np.array([True, True, False]))

    def test_from_range(self):

        comp = CategoricalComponent(np.array(list('abcdefghijklmnopqrstuvwxyz') * 2))

        roi = CategoricalROI.from_range(comp, 6, 10)
        np.testing.assert_array_equal(roi.categories,
                                      np.array(list('ghij')))

    def test_empty_categories(self):
        roi = CategoricalROI()
        contains = roi.contains(np.array(['a', 'b', 'c']), None)
        np.testing.assert_array_equal(contains, [0, 0, 0])


class DummyEvent(object):
    def __init__(self, x, y, inaxes=True, key=None):
        self.inaxes = inaxes
        self.xdata = x
        self.ydata = y
        self.key = key


class MockAxes(object):
    def __init__(self):
        self.figure = MagicMock()
        self.figure.canvas = MagicMock()

    def add_patch(self, patch):
        pass


class TestMpl(object):

    def setup_method(self, method):
        self.axes = MagicMock()
        self.roi = self._roi_factory()

    def _roi_factory(self):
        raise NotImplementedError

    def test_undefined_on_creation(self):
        assert not self.roi._roi.defined()

    def test_large_zorder(self):
        assert self.roi._patch.get_zorder() >= 100

    def test_proper_roi(self):
        raise NotImplementedError

    def test_start_ignored_if_not_inaxes(self):
        ev = DummyEvent(0, 0, inaxes=None)
        self.roi.start_selection(ev)
        assert not self.roi._roi.defined()

    def test_canvas_syncs_properly(self):
        assert self.axes.figure.canvas.draw_idle.call_count == 0
        assert self.axes.figure.canvas.blit.call_count == 0
        event = DummyEvent(5, 5, inaxes=self.axes)
        self.roi.start_selection(event)
        assert self.axes.figure.canvas.draw_idle.call_count == 0
        assert self.axes.figure.canvas.blit.call_count == 1
        self.roi.update_selection(event)
        assert self.axes.figure.canvas.draw_idle.call_count == 0
        assert self.axes.figure.canvas.blit.call_count == 2
        self.roi.update_selection(event)
        assert self.axes.figure.canvas.draw_idle.call_count == 0
        assert self.axes.figure.canvas.blit.call_count == 3
        self.roi.finalize_selection(event)
        assert self.axes.figure.canvas.draw_idle.call_count == 0
        assert self.axes.figure.canvas.blit.call_count == 4

    def test_patch_shown_on_start(self):
        assert not self.roi._patch.get_visible()
        event = DummyEvent(5, 5, inaxes=self.axes)
        self.roi.start_selection(event)
        assert self.roi._patch.get_visible()

    def test_patch_hidden_on_finalise(self):
        event = DummyEvent(5, 5, inaxes=self.axes)
        self.roi.start_selection(event)
        self.roi.finalize_selection(event)
        assert not self.roi._patch.get_visible()

    def test_update_before_start_ignored(self):
        self.roi.update_selection(None)
        assert not self.roi._roi.defined()

    def test_finalize_before_start_ignored(self):
        self.roi.finalize_selection(None)
        assert not self.roi._roi.defined()

    def test_roi_defined_after_start(self):
        event = DummyEvent(5, 5, inaxes=self.axes)
        self.roi.start_selection(event)
        assert self.roi._roi.defined()

    def test_roi_undefined_before_start(self):
        assert not self.roi._roi.defined()

    def scrub(self, roi=None, abort=False, outside=False):

        if roi is None:

            roi = self._roi_factory()

            event = DummyEvent(5, 5, inaxes=self.axes)
            roi.start_selection(event)
            event = DummyEvent(10, 10, inaxes=self.axes)
            roi.update_selection(event)
            roi.finalize_selection(DummyEvent(10, 10))

        if outside:
            roi.start_selection(DummyEvent(16, 16, inaxes=self.axes, key='control'))
            roi.update_selection(DummyEvent(17, 18, inaxes=self.axes, key='control'))
        else:
            roi.start_selection(DummyEvent(6, 6, inaxes=self.axes, key='control'))
            roi.update_selection(DummyEvent(7, 8, inaxes=self.axes, key='control'))

        if abort:
            roi.abort_selection(None)

        return roi


class TestRectangleMpl(TestMpl):

    def _roi_factory(self):
        return MplRectangularROI(self.axes)

    def test_scrub(self):

        roi = self.scrub()

        assert roi._roi.xmin == 6
        assert roi._roi.xmax == 11
        assert roi._roi.ymin == 7
        assert roi._roi.ymax == 12

    def test_abort(self):

        roi = self.scrub(abort=True)

        assert roi._roi.xmin == 5
        assert roi._roi.xmax == 10
        assert roi._roi.ymin == 5
        assert roi._roi.ymax == 10

    def test_outside(self):

        roi = self.scrub(outside=True)

        assert roi._roi.xmin == 5
        assert roi._roi.xmax == 10
        assert roi._roi.ymin == 5
        assert roi._roi.ymax == 10

    def assert_roi_correct(self, x0, x1, y0, y1):
        corner = self.roi.roi().corner()
        height = self.roi.roi().height()
        width = self.roi.roi().width()
        assert self.roi.roi().defined()
        assert_almost_equal(corner[0], min(x0, x1), 4)
        assert_almost_equal(corner[1], min(y0, y1), 4)
        assert_almost_equal(abs(y1 - y0), height, 4)
        assert_almost_equal(abs(x1 - x0), width, 4)

    def assert_patch_correct(self, x0, x1, y0, y1):
        corner = self.roi._patch.get_xy()
        width = self.roi._patch.get_width()
        height = self.roi._patch.get_height()
        assert_almost_equal(corner, (min(x0, x1), min(y0, y1)), 4)
        assert_almost_equal(width, abs(x0 - x1))
        assert_almost_equal(height, abs(y0 - y1))

    def test_str(self):
        assert type(str(self.roi)) == str

    def test_proper_roi(self):
        assert isinstance(self.roi._roi, RectangularROI)

    def test_roi_on_start_selection(self):
        event = DummyEvent(5, 5, inaxes=self.axes)
        self.roi.start_selection(event)
        self.assert_roi_correct(5, 5, 5, 5)

    def test_patch_on_start_selection(self):
        event = DummyEvent(5, 5, inaxes=self.axes)
        self.roi.start_selection(event)
        self.assert_patch_correct(5, 5, 5, 5)

    def test_update_selection(self):
        event = DummyEvent(5, 6, inaxes=self.axes)
        self.roi.start_selection(event)
        event = DummyEvent(10, 11, inaxes=self.axes)
        self.roi.update_selection(event)
        self.assert_roi_correct(5, 10, 6, 11)
        self.assert_patch_correct(5, 10, 6, 11)

    def test_finalize_selection(self):
        event = DummyEvent(5, 6, inaxes=self.axes)
        self.roi.start_selection(event)
        event = DummyEvent(10, 11, inaxes=self.axes)
        self.roi.update_selection(event)
        self.roi.finalize_selection(event)
        self.assert_roi_correct(5, 10, 6, 11)
        self.assert_patch_correct(5, 10, 6, 11)

    def test_define_roi_to_right(self):
        ev0 = DummyEvent(0.5, 0.5, inaxes=self.axes)
        ev1 = DummyEvent(1, 1, inaxes=self.axes)
        self.roi.start_selection(ev0)
        self.roi.update_selection(ev1)
        self.roi.finalize_selection(ev1)

        self.assert_roi_correct(.5, 1, .5, 1)
        self.assert_patch_correct(.5, 1, .5, 1)

    def test_define_roi_to_left(self):
        ev0 = DummyEvent(1, 1, inaxes=self.axes)
        ev1 = DummyEvent(0.5, 0.5, inaxes=self.axes)
        self.roi.start_selection(ev0)
        self.roi.update_selection(ev1)
        self.roi.finalize_selection(ev1)

        self.assert_roi_correct(.5, 1, .5, 1)
        self.assert_patch_correct(.5, 1, .5, 1)


class TestXRangeMpl(TestMpl):
    def _roi_factory(self):
        return MplXRangeROI(self.axes)

    def test_proper_roi(self):
        assert isinstance(self.roi._roi, XRangeROI)

    def test_start_selection(self):
        event = DummyEvent(1, 1, inaxes=self.axes)
        self.roi.start_selection(event)
        assert self.roi._roi.defined()

    def test_update_selection(self):
        ev0 = DummyEvent(1, 1, inaxes=self.axes)
        ev1 = DummyEvent(2, 1, inaxes=self.axes)
        self.roi.start_selection(ev0)
        self.roi.update_selection(ev1)
        assert self.roi._roi.defined()
        assert self.roi._roi.range() == (1, 2)

    def test_finalize_selection(self):
        ev0 = DummyEvent(1, 1, inaxes=self.axes)
        ev1 = DummyEvent(2, 1, inaxes=self.axes)
        self.roi.start_selection(ev0)
        self.roi.update_selection(ev1)
        self.roi.finalize_selection(ev1)
        assert self.roi._roi.defined()
        assert self.roi._roi.range() == (1, 2)

        assert not self.roi._patch.get_visible()

    def test_scrub(self):

        roi = self.scrub()

        assert_almost_equal(roi._roi.min, 3.0)
        assert_almost_equal(roi._roi.max, 8.0)

    def test_abort(self):

        roi = self.scrub(abort=True)

        assert_almost_equal(roi._roi.min, 5.0)
        assert_almost_equal(roi._roi.max, 10.0)

    def test_outside(self):

        roi = self.scrub(outside=True)

        assert_almost_equal(roi._roi.min, 5.0)
        assert_almost_equal(roi._roi.max, 10.0)


class TestYRangeMpl(TestMpl):
    def _roi_factory(self):
        return MplYRangeROI(self.axes)

    def test_proper_roi(self):
        assert isinstance(self.roi._roi, YRangeROI)

    def test_start_selection(self):
        event = DummyEvent(1, 1, inaxes=self.axes)
        self.roi.start_selection(event)
        assert self.roi._roi.defined()

    def test_update_selection(self):
        ev0 = DummyEvent(1, 1, inaxes=self.axes)
        ev1 = DummyEvent(1, 2, inaxes=self.axes)
        self.roi.start_selection(ev0)
        self.roi.update_selection(ev1)
        assert self.roi._roi.defined()
        assert self.roi._roi.range() == (1, 2)

    def test_finalize_selection(self):
        ev0 = DummyEvent(1, 1, inaxes=self.axes)
        ev1 = DummyEvent(1, 2, inaxes=self.axes)
        self.roi.start_selection(ev0)
        self.roi.update_selection(ev1)
        self.roi.finalize_selection(ev1)
        assert self.roi._roi.defined()
        assert self.roi._roi.range() == (1, 2)

        assert not self.roi._patch.get_visible()

    def test_scrub(self):

        roi = self.scrub()

        assert_almost_equal(roi._roi.min, 4.0)
        assert_almost_equal(roi._roi.max, 9.0)

    def test_abort(self):

        roi = self.scrub(abort=True)

        assert_almost_equal(roi._roi.min, 5.0)
        assert_almost_equal(roi._roi.max, 10.0)

    def test_outside(self):

        roi = self.scrub(outside=True)

        assert_almost_equal(roi._roi.min, 5.0)
        assert_almost_equal(roi._roi.max, 10.0)


def test_circular_roi_representation():

    # Test cases where drawn circular ROIs are converted to circles, ellipses,
    # and polygons. Here we don't override pixel_to_data and data_to_pixel
    # since these are important.

    event = namedtuple('Event', ['inaxes', 'xdata', 'ydata', 'key'])

    fig = plt.figure(figsize=(6, 4))
    ax = fig.add_subplot(1, 1, 1)

    # Case 1: log-linear axes

    ax.set_xlim(10, 100)
    ax.set_ylim(-2, 5)
    ax.set_xscale('log')
    ax.set_yscale('linear')

    fig.canvas.draw()

    mpl_roi = MplCircularROI(ax)
    mpl_roi.start_selection(event(inaxes=ax, xdata=50, ydata=1, key=None))
    mpl_roi.update_selection(event(inaxes=ax, xdata=100, ydata=1, key=None))
    roi = mpl_roi.roi()
    assert isinstance(roi, PolygonalROI)

    # Case 2: linear-linear axes with non-square aspect ratio

    ax.set_xlim(-40, 100)
    ax.set_ylim(-2, 5)
    ax.set_xscale('linear')
    ax.set_yscale('linear')

    fig.canvas.draw()

    mpl_roi = MplCircularROI(ax)
    mpl_roi.start_selection(event(inaxes=ax, xdata=30, ydata=1, key=None))
    mpl_roi.update_selection(event(inaxes=ax, xdata=80, ydata=1, key=None))
    roi = mpl_roi.roi()
    assert isinstance(roi, EllipticalROI)
    assert_allclose(roi.xc, 30)
    assert_allclose(roi.yc, 1)
    assert_allclose(roi.radius_x, 50)
    assert_allclose(roi.radius_y, 5.871212)

    # Case 3: linear-linear axes with square aspect ratio

    ax.set_aspect('equal')
    ax.set_xlim(-40, 100)

    fig.canvas.draw()

    mpl_roi = MplCircularROI(ax)
    mpl_roi.start_selection(event(inaxes=ax, xdata=50, ydata=1, key=None))
    mpl_roi.update_selection(event(inaxes=ax, xdata=80, ydata=1, key=None))
    roi = mpl_roi.roi()
    assert isinstance(roi, CircularROI)
    assert_allclose(roi.xc, 50)
    assert_allclose(roi.yc, 1)
    assert_allclose(roi.radius, 30)


class TestCircleMpl(TestMpl):

    def _roi_factory(self):
        return MplCircularROI(self.axes)

    def setup_method(self, method):
        super(TestCircleMpl, self).setup_method(method)
        self.pixel_to_data = r.pixel_to_data
        self.data_to_pixel = r.data_to_pixel

        r.pixel_to_data = lambda x, y, z: np.column_stack((y, z))
        r.data_to_pixel = lambda x, y, z: np.column_stack((y, z))

    def teardown_method(self, method):
        # restore methods
        r.pixel_to_data = self.pixel_to_data
        r.data_to_pixel = self.data_to_pixel

    def test_proper_roi(self):
        assert isinstance(self.roi._roi, CircularROI)

    def test_to_polygon_undefined(self):
        """to_polygon() result should be undefined before defining polygon"""
        roi = self.roi.roi()
        assert not roi.defined()

    def test_roi_defined_correctly(self):
        ev0 = DummyEvent(0, 0, inaxes=self.axes)
        ev1 = DummyEvent(5, 0, inaxes=self.axes)
        self.roi.start_selection(ev0)
        self.roi.update_selection(ev1)
        self.roi.finalize_selection(ev1)
        self.assert_roi_correct(0, 0, 5)

    def assert_roi_correct(self, x, y, r):
        roi = self.roi.roi()
        assert roi.defined()
        assert roi.contains(x, y)
        assert roi.contains(x + .95 * r, y)
        assert not roi.contains(x + 1.05 * r, y)
        assert not roi.contains(x + .8 * r, y + .8 * r)

    def test_scrub(self):

        roi = self.scrub()

        assert roi._roi.xc == 6
        assert roi._roi.yc == 7
        assert_almost_equal(roi._roi.radius, 7.0, decimal=0)

    def test_abort(self):

        roi = self.scrub(abort=True)

        assert roi._roi.xc == 5
        assert roi._roi.yc == 5
        assert_almost_equal(roi._roi.radius, 7.0, decimal=0)

    def test_outside(self):

        roi = self.scrub(outside=True)

        assert roi._roi.xc == 5
        assert roi._roi.yc == 5
        assert_almost_equal(roi._roi.radius, 7.0, decimal=0)


class TestPolyMpl(TestMpl):
    def _roi_factory(self):
        return MplPolygonalROI(self.axes)

    def test_proper_roi(self):
        return isinstance(self.roi._roi, PolygonalROI)

    def send_events(self):
        ev0 = DummyEvent(5, 5, inaxes=self.axes)
        ev1 = DummyEvent(5, 10, inaxes=self.axes)
        ev2 = DummyEvent(10, 10, inaxes=self.axes)
        ev3 = DummyEvent(10, 5, inaxes=self.axes)
        self.roi.start_selection(ev0)
        self.roi.update_selection(ev1)
        self.roi.update_selection(ev2)
        self.roi.update_selection(ev3)
        self.roi.finalize_selection(ev3)

    def assert_roi_correct(self):
        roi = self.roi.roi()
        assert roi.contains(7.0, 7.0)
        assert not roi.contains(12, 7.0)

    def test_define(self):
        self.send_events()
        self.assert_roi_correct()

    def test_scrub(self):

        self.send_events()

        roi = self.scrub(roi=self.roi)

        assert roi._roi.vx[0] == 6
        assert roi._roi.vy[0] == 7

    def test_abort(self):

        self.send_events()

        roi = self.scrub(roi=self.roi, abort=True)

        assert roi._roi.vx[0] == 5
        assert roi._roi.vy[0] == 5

    def test_outside(self):

        self.send_events()

        roi = self.scrub(roi=self.roi, outside=True)

        assert roi._roi.vx[0] == 5
        assert roi._roi.vy[0] == 5


class TestPickMpl(TestMpl):
    def _roi_factory(self):
        return MplPickROI(self.axes)

    def test_proper_roi(self):
        return isinstance(self.roi._roi, PointROI)

    def test_start_ignored_if_not_inaxes(self):
        ev = DummyEvent(0, 0, inaxes=None)
        self.roi.start_selection(ev)
        assert self.roi._roi.defined()

    def test_update_before_start_ignored(self):
        ev = DummyEvent(0, 0, inaxes=None)
        self.roi.update_selection(ev)
        assert self.roi._roi.defined()

    def test_finalize_before_start_ignored(self):
        ev = DummyEvent(0, 0, inaxes=None)
        self.roi.finalize_selection(ev)
        assert self.roi._roi.defined()

    def test_large_zorder(self):
        """No patch to test for."""

    def test_patch_shown_on_start(self):
        """No patch to test for."""

    def test_patch_hidden_on_finalise(self):
        """No patch to test for."""

    def test_canvas_syncs_properly(self):
        """No patch to test for."""


class TestUtil(object):

    def setup_method(self, method):
        self.axes = AXES

    def test_aspect_ratio(self):
        self.axes.figure.set_figheight(5)
        self.axes.figure.set_figwidth(5)
        self.axes.set_ylim([0, 10])
        self.axes.set_xlim([0, 10])

        ax0 = r.aspect_ratio(self.axes)
        self.axes.set_ylim(0, 20)
        assert_almost_equal(r.aspect_ratio(self.axes), ax0 / 2, 4)
        self.axes.set_ylim(0, 5)
        assert_almost_equal(r.aspect_ratio(self.axes), ax0 * 2, 4)
        self.axes.set_xlim(0, 5)
        assert_almost_equal(r.aspect_ratio(self.axes), ax0, 4)
        self.axes.set_xlim(0, 10)
        assert_almost_equal(r.aspect_ratio(self.axes), ax0 * 2, 4)

    def test_data_to_norm_with_scalars(self):
        # assume data that gets submitted to axes is valid.
        # testing to see if we can get there
        self.axes.set_xlim(0, 10)
        self.axes.set_ylim(0, 10)
        func = r.data_to_norm
        assert_almost_equal(func(self.axes, 0, 0)[0, 0], 0, 3)
        assert_almost_equal(func(self.axes, 0, 0)[0, 1], 0, 3)
        assert_almost_equal(func(self.axes, 0, 10)[0, 0], 0, 3)
        assert_almost_equal(func(self.axes, 0, 10)[0, 1], 1, 3)
        assert_almost_equal(func(self.axes, 10, 10)[0, 0], 1, 3)
        assert_almost_equal(func(self.axes, 10, 10)[0, 1], 1, 3)
        assert_almost_equal(func(self.axes, 10, 0)[0, 0], 1, 3)
        assert_almost_equal(func(self.axes, 10, 0)[0, 1], 0, 3)

        x = np.array([0, 0, 10, 10])
        y = np.array([0, 10, 0, 10])
        xans = [0, 0, 1, 1]
        yans = [0, 1, 0, 1]
        norm = func(self.axes, x, y)
        assert_almost_equal(norm[0, 0], xans[0], 3)
        assert_almost_equal(norm[1, 0], xans[1], 3)
        assert_almost_equal(norm[2, 0], xans[2], 3)
        assert_almost_equal(norm[3, 0], xans[3], 3)
        assert_almost_equal(norm[0, 1], yans[0], 3)
        assert_almost_equal(norm[1, 1], yans[1], 3)
        assert_almost_equal(norm[2, 1], yans[2], 3)
        assert_almost_equal(norm[3, 1], yans[3], 3)

        x = [0, 0, 10, 10]
        y = [0, 10, 0, 10]
        assert_almost_equal(norm[0, 0], xans[0], 3)
        assert_almost_equal(norm[1, 0], xans[1], 3)
        assert_almost_equal(norm[2, 0], xans[2], 3)
        assert_almost_equal(norm[3, 0], xans[3], 3)
        assert_almost_equal(norm[0, 1], yans[0], 3)
        assert_almost_equal(norm[1, 1], yans[1], 3)
        assert_almost_equal(norm[2, 1], yans[2], 3)
        assert_almost_equal(norm[3, 1], yans[3], 3)

    def test_data_to_pixel(self):
        xp = 100
        yp = 100
        data = r.pixel_to_data(self.axes, xp, yp)
        pixel = r.data_to_pixel(self.axes, data[:, 0], data[:, 1])
        assert_almost_equal(pixel[0, 0], xp, 3)
        assert_almost_equal(pixel[0, 1], yp, 3)


del TestMpl  # prevents unittest discovery from running abstract base class



from __future__ import print_function

import numpy as np

from scipy.ndimage import map_coordinates


def extract_line_slice(cube, x, y, order=3, respect_nan=True):
    """
    Given an array with shape (z, y, x), extract a (z, n) slice by
    interpolating at n (x, y) points.

    All units are in *pixels*.

    .. note:: If there are NaNs in the cube, they will be treated as zeros when
              using spline interpolation.

    Parameters
    ----------
    cube : `~numpy.ndarray`
        The data cube to extract the slice from
    curve : list or tuple
        A list or tuple of (x, y) pairs, with minimum length 2
    order : int, optional
        Spline interpolation order. Set to ``0`` for nearest-neighbor
        interpolation.

    Returns
    -------
    slice : `numpy.ndarray`
        The (z, d) slice
    """

    if order == 0:

        total_slice = np.zeros([cube.shape[0], len(x)]) + np.nan

        x = np.round(x)
        y = np.round(y)

        ok = (x >= 0) & (y >= 0) & (x < cube.shape[2]) & (y < cube.shape[1])

        total_slice[:,ok] = cube[:, y[ok].astype(int), x[ok].astype(int)]

    elif order > 0 and order == int(order):

        nx = len(x)
        nz = cube.shape[0]

        zi = np.outer(np.arange(nz, dtype=int), np.ones(nx))
        xi = np.outer(np.ones(nz), x)
        yi = np.outer(np.ones(nz), y)

        if np.any(np.isnan(cube)):

            # map_coordinates does not deal well with NaN values so we have
            # to remove the NaN values then re-mask the final slice.

            total_slice = map_coordinates(np.nan_to_num(cube), [zi,yi,xi],
                                          order=order, cval=np.nan)

            slice_bad = map_coordinates(np.nan_to_num(np.isnan(cube).astype(int)),
                                        [zi,yi,xi], order=order)

            total_slice[np.nonzero(slice_bad)] = np.nan

        else:

            total_slice = map_coordinates(cube, [zi,yi,xi], order=order, cval=np.nan)

    else:

        raise TypeError("order should be a positive integer")

    return total_slice
